package nasa.stps;
import nasa.stps.core.fs.FrameSynchronizer;
import nasa.stps.core.Builder;
import nasa.stps.core.StpsException;
import nasa.stps.core.Configuration;
import nasa.stps.core.StpsNode;
import nasa.stps.core.status.StatusItem;
import org.xml.sax.InputSource;
import java.io.File;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.FileReader;
import java.io.BufferedReader;

/**
 * This class is the main entry for the STPS Batch Processor. Batch is a
 * standalone, comand-line-invoked program that reads a raw telemetry file
 * and produces one or more files of packets, frames, and/or other units.
 * @author Stan Hilinski, April 2001
 */
public final class StpsBatch
{
    private static final int BUFFERLENGTH = 8192;

    public static void main(String[] args)
    {
        System.out.println(nasa.stps.Version.getVersion());

        if (args.length != 2)
        {
            System.err.println("Options: [-Droot=<stp.dtdDirectory>] [-Dsetup=<directory>] [-Draw=<directory>]");
            System.err.println("Arguments: [setupFile] [dataFile]");
            System.exit(1);
        }

        FrameSynchronizer fs = null;
        Configuration configuration = null;

        try
        {
            String directory = System.getProperty("setup");
            File setupFile = new File(directory,args[0]);
            BufferedReader br = new BufferedReader(new FileReader(setupFile));
            InputSource setup = new InputSource(br);
            String root = System.getProperty("root","file://localhost/");
            setup.setSystemId(root);

            Builder builder = new Builder();
            fs = builder.create(setup);
            configuration = builder.getConfiguration();
            br.close();
        }
        catch (java.io.IOException ifnf)
        {
            System.err.println(ifnf.getMessage());
            System.exit(1);
        }
        catch (StpsException be)
        {
            System.err.println(be.getMessage());
            System.exit(2);
        }

        BufferedInputStream input = null;
        try
        {
            String directory = System.getProperty("raw");
            File dataFile = new File(directory,args[1]);
            FileInputStream fis = new FileInputStream(dataFile);
            input = new BufferedInputStream(fis);
        }
        catch (java.io.FileNotFoundException ofnf)
        {
            ofnf.printStackTrace();
            System.exit(1);
        }

        long tbytes = 0L;
        long t0 = System.currentTimeMillis();

        byte[] data = new byte[BUFFERLENGTH];

        try
        {
            while (true)
            {
                int bytes = input.read(data,0,BUFFERLENGTH);
                if (bytes == -1) break;
                fs.putBuffer(data,bytes);
                tbytes += bytes;
            }

            fs.shutdown();
        }
        catch (java.io.IOException re)
        {
            re.printStackTrace();
            System.exit(1);
        }
        catch (StpsException stpe)
        {
            stpe.printStackTrace();
            System.exit(2);
        }

        long t1 = System.currentTimeMillis();
        long t = t1 - t0;
        long bitsPerMs = 8L * tbytes / t;
        long bitsPerSecond = bitsPerMs * 1000L;
        long kilobitsPerSecond = bitsPerSecond / 1024L;
        System.out.println("kilobits per second="+ kilobitsPerSecond);
        System.out.println();

        try
        {
            input.close();
        }
        catch (java.io.IOException ce)
        {
            ce.printStackTrace();
        }

        java.util.Map nodes = configuration.getStpsNodes();
        java.util.Iterator i = nodes.values().iterator();
        while (i.hasNext())
        {
            StpsNode node = (StpsNode)i.next();
            java.util.Collection status = node.getStatusItems();
            if (status != null)
            {
                System.out.println(node);
                java.util.Iterator ii = status.iterator();
                while (ii.hasNext())
                {
                    StatusItem si = (StatusItem)ii.next();
                    System.out.println("    " + si);
                }
            }
        }
    }
}
