package nasa.stps.core;

import nasa.stps.core.fs.*;
import nasa.stps.core.ccsds.*;
import nasa.stps.core.ccsds.path.*;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * This class creates an STPS pipeline.
 * @author Stan Hilinski, March, 2001
 */
public class Builder
{
    private StpsNodeFactory nodeFactory = new StpsNodeFactory();
    private DocumentBuilder documentBuilder = null;
    private Configuration config = null;

    public Builder() throws StpsException
    {
        //I create my node factory.
        populateNodeFactory();

        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        dbf.setValidating(true);
        dbf.setIgnoringComments(true);
        dbf.setIgnoringElementContentWhitespace(true);

        try
        {
            documentBuilder = dbf.newDocumentBuilder();
            documentBuilder.setErrorHandler(new MyErrorHandler());
        }
        catch (ParserConfigurationException pce)
        {
            throw new StpsException(pce);
        }
    }

    /**
     * Get the STPS node factory.
     */
    public final StpsNodeFactory getStpsNodeFactory()
    {
        return nodeFactory;
    }

    /**
     * Get the last-used configuration. The builder does not change the
     * configuration until after it creates an STPS pipeline. It will be
     * null if no configuration has ever been loaded.
     */
    public final Configuration getConfiguration()
    {
        return config;
    }

    /**
     * Create an STPS pipeline.
     * @return The FrameSynchronizer object.
     */
    public FrameSynchronizer create(org.xml.sax.InputSource input)
            throws StpsException
    {
        //I build an XML DOM document from the loaded configuration file.
        Document document = null;
        try
        {
            document = documentBuilder.parse(input);
        }
        catch (SAXException se)
        {
            throw new StpsException(se);
        }
        catch (java.io.IOException ioe)
        {
            throw new StpsException(ioe);
        }

        //The root element is the top of the document.
        Element root = document.getDocumentElement();
        if (!root.getTagName().equals("stps"))
        {
            throw new StpsException("The setup is not an stps configuration.");
        }

        //The cid is a description of the configuration, which I made as a
        //root element attribute.
        String cid = root.getAttribute("id");

        //I use the configuration object to store shared information.
        config = new Configuration(cid,nodeFactory);

         //I do the spacecraft first so that stps nodes that need them may
         //map them immediately.
        createSpacecraftMap(document,config);

         //Create all STPS nodes. I save the links element to process later.
        Element links = createNodes(root,config);

         //Link all nodes. These are all links defined by link statements.
         //It does not include special links as defined in the CADU and
         //Path nodes.
        linkNodes(links,config);

         //Allow each node to finish setup. Nodes can now assume that all
         //other nodes have been created. Some nodes will resolve their
         //special links, while others will get shared information from
         //other nodes.
        java.util.Iterator i = config.getStpsNodes().values().iterator();
        while (i.hasNext())
        {
            StpsNode node = (StpsNode)i.next();
            node.finishSetup(config);
        }

        //There is one FS, so the node name is the same as the class name.
        String fsNodeName = FrameSynchronizer.CLASSNAME;
        return (FrameSynchronizer)config.getStpsNodes().get(fsNodeName);
    }

    private void populateNodeFactory() throws StpsException
    {
        nodeFactory.addNode(FrameSynchronizer.CLASSNAME,new FrameSynchronizer());
        nodeFactory.addNode(CrcDecoder.CLASSNAME,new CrcDecoder());
        nodeFactory.addNode(PnDecoder.CLASSNAME,new PnDecoder());
        nodeFactory.addNode(ReedSolomonDecoder.CLASSNAME,new ReedSolomonDecoder());
        nodeFactory.addNode(FrameStatus.CLASSNAME,new FrameStatus());

        nodeFactory.addNode(CaduService.CLASSNAME,new CaduService());
        nodeFactory.addNode(VcduService.CLASSNAME,new VcduService());
        nodeFactory.addNode(PathService.CLASSNAME,new PathService());
        nodeFactory.addNode(PacketPipeline.CLASSNAME,new PacketPipeline());

        nodeFactory.addNode(nasa.stps.core.xstps.hrit.HRITOutput.CLASSNAME,
                new nasa.stps.core.xstps.hrit.HRITOutput(this));
    }

    /**
     * Ingest all spacecrafts from the xml configuration file.
     */
    private void createSpacecraftMap(Document document,
            Configuration config) throws StpsException
    {
        NodeList list = document.getElementsByTagName("spacecraft");
        int length = list.getLength();
        if (length > 0)
        {
            java.util.Map spacecrafts = config.getSpacecrafts();
            for (int n = 0; n < length; n++)
            {
                Spacecraft s = new Spacecraft((Element)list.item(n));
                spacecrafts.put(s.getName(),s);
            }
        }
    }

    private Element createNodes(Element root, Configuration config)
            throws StpsException
    {
        Element links = null;
        java.util.Map stpsNodes = config.getStpsNodes();

        /**
         * I create the FrameStatus node here because there is no element
         * in the setup file related to it.
         */
        StpsNode frameStatus = nodeFactory.create("frame_status");
        frameStatus.load(null,config);
        stpsNodes.put(frameStatus.getLinkName(),frameStatus);

         //The PN decoder does not have its own element, so it is not
         //created here. Since the pn flag is in FsSetup, the
         //FrameSynchronizer node creates PnDecoder when FrameSynchronizer
         //is loaded.
        NodeList list = root.getChildNodes();

        int length = list.getLength();
        for (int n = 0; n < length; n++)
        {
            Element element = (Element)list.item(n);
            String id = element.getTagName();

            if (id.equals("links"))
            {
                //Save the links element for later to ensure that I've done
                //all nodes first.
                links = element;
            }
            else if (id.equals("spacecrafts"))
            {
                continue;  //spacecraft were already completed
            }
            else if (id.equals("ccsds_services"))
            {
                createServiceNodes(element,config);
            }
            else if (id.equals("packets"))
            {
                createPacketNodes(element,config);
            }
            else if (id.equals("output_channels"))
            {
                createOutputChannels(element,config);
            }
            else
            {
                StpsNode snode = nodeFactory.create(element,config);
                if (snode == null)
                {
                    throw new StpsException("Unknown node type " +
                        element.getTagName());
                }
                if (snode.getLinkName() == null)
                {
                    throw new StpsException(element.getTagName() +
                        " does not have a link name.");
                }
                stpsNodes.put(snode.getLinkName(),snode);
            }
        }

        return links;
    }

    private void createServiceNodes(Element element, Configuration config)
            throws StpsException
    {
        NodeList list = element.getChildNodes();
        int length = list.getLength();
        java.util.Map stpsNodes = config.getStpsNodes();
        StpsNodeFactory nodeFactory = config.getNodeFactory();
        for (int s = 0; s < length; s++)
        {
            element = (Element)list.item(s);
            StpsNode service = nodeFactory.create(element,config);
            stpsNodes.put(service.getLinkName(),service);
        }
    }

    private void createPacketNodes(Element element, Configuration config)
            throws StpsException
    {
        NodeList list = element.getElementsByTagName("packet");
        int length = list.getLength();
        StpsNodeFactory nodeFactory = config.getNodeFactory();
        java.util.Map stpsNodes = config.getStpsNodes();
        for (int p = 0; p < length; p++)
        {
            element = (Element)list.item(p);
            StpsNode packet = nodeFactory.create(element,config);
            stpsNodes.put(packet.getLinkName(),packet);
        }
    }

    private void createOutputChannels(Element element, Configuration config)
            throws StpsException
    {
        NodeList list = element.getChildNodes();
        int length = list.getLength();
        StpsNodeFactory nodeFactory = config.getNodeFactory();
        java.util.Map stpsNodes = config.getStpsNodes();
        for (int p = 0; p < length; p++)
        {
            Node xnode = list.item(p);
            if (xnode instanceof Element)
            {
                element = (Element)list.item(p);
                String id = element.getTagName();
                if (id.equals("file") || id.equals("socket"))
                {
                    String utype = element.getAttribute("unitType");
                    String annotation = element.getAttribute("annotation");
                    if (annotation.length() == 0) annotation = "BEFORE";
                    id = utype + '_' + annotation;
                    StpsNode oo = nodeFactory.create(id);
                    oo.load(element,config);
                    stpsNodes.put(oo.getLinkName(),oo);
                }
                else
                {
                    StpsNode oo = nodeFactory.create(element,config);
                    stpsNodes.put(oo.getLinkName(),oo);
                }
            }
            else
            {
                throw new StpsException("Invalid output element in configuration.");
            }
        }
    }

    private void linkNodes(Element element, Configuration config)
            throws StpsException
    {
        java.util.Map stpsNodes = config.getStpsNodes();
        NodeList links = element.getChildNodes();
        int length = links.getLength();
        for (int n = 0; n < length; n++)
        {
            Element link = (Element)links.item(n);

            String from = link.getAttribute("from");
            StpsNode source = (StpsNode)stpsNodes.get(from);
            if (source == null)
            {
                throw new StpsException("Source node " + from +
                        " does not exist.");
            }

            String to = link.getAttribute("to");
            StpsNode destination = (StpsNode)stpsNodes.get(to);
            if (destination == null)
            {
                throw new StpsException("Destination node " + to +
                        " does not exist.");
            }

            if (source instanceof Sender)
            {
                if (destination instanceof Receiver)
                {
                    Sender sender = (Sender)source;
                    sender.addReceiver((Receiver)destination);
                }
                else
                {
                    throw new StpsException("Destination " + to +
                            " is not a receiver.");
                }
            }
            else
            {
                throw new StpsException("Source " + from +
                        " is not a sender.");
            }
        }
    }

    class MyErrorHandler implements org.xml.sax.ErrorHandler
    {
        /**
         * Returns a string describing parse exception details
         */
        private String getParseExceptionInfo(SAXParseException spe)
        {
            String systemId = spe.getSystemId();
            if (systemId == null) systemId = "null";
            String info = "URI=" + systemId + " Line=" + spe.getLineNumber() +
                ": " + spe.getMessage();
            return info;
        }

        public void warning(SAXParseException spe) throws SAXException
        {
            System.err.println("Warning: " + getParseExceptionInfo(spe));
        }

        public void error(SAXParseException spe) throws SAXException
        {
            String message = "Error: " + getParseExceptionInfo(spe);
            throw new SAXException(message);
        }

        public void fatalError(SAXParseException spe) throws SAXException
        {
            String message = "Fatal Error: " + getParseExceptionInfo(spe);
            throw new SAXException(message);
        }
    }
}