package nasa.stps.core;

/**
 * This class is an STPS node factory. It maps unique node ids to blank
 * instances of nodes. The node id is the element tag in the xml
 * configuration file and is a class variable. (It is not the same as
 * a link name, which is the unique name given to one instance of a node.)
 * To create a node, pass an id, and the factory returns a clone of the
 * mapped node.
 * @author Stan Hilinski, May 2001
 */
public class StpsNodeFactory extends java.util.TreeMap
{
    /**
     * Add an STPS node (StpsNode) object to the factory. This node is an
     * empty template and should have significant state because it will be
     * cloned to create useable instances.
     * @param id A node type name, which often matches a element tag name.
     *          It must be unique.
     * @param node The STPS node
     */
    public void addNode(String id, StpsNode node) throws StpsException
    {
        Object duplicate = put(id,node);
        if (duplicate != null)
        {
            throw new StpsException(id + " is not a unique stps node name.");
        }
    }

    /**
     * Create an StpsNode.
     * @param id A node type name.
     * @return An StpsNode. It is not initialized, so you must load it.
     */
    public StpsNode create(String id) throws StpsException
    {
        StpsNode node = (StpsNode)get(id);
        if (node == null)
        {
            throw new StpsException(id + " is an unknown node type.");
        }

        try
        {
            node = (StpsNode)node.clone();
        }
        catch (CloneNotSupportedException cnse)
        {
            throw new StpsException(id + " class is not cloneable.");
        }

        return node;
    }

    /**
     * Create an StpsNode.
     * @param element An XML element from the STPS configuration file.
     *          The element tag name is the STPS Node id.
     * @param config A collection of configuration items.
     * @return An StpsNode. It is initialized.
     */
    public StpsNode create(org.w3c.dom.Element element, Configuration config)
            throws StpsException
    {
        String id = element.getTagName();
        StpsNode node = (StpsNode)get(id);
        if (node == null)
        {
            throw new StpsException(id + " is an unknown node type.");
        }

        try
        {
            node = (StpsNode)node.clone();
        }
        catch (CloneNotSupportedException cnse)
        {
            throw new StpsException(id + " class is not cloneable.");
        }

        node.load(element,config);

        return node;
    }
}