package nasa.stps.core.ccsds;

import nasa.stps.core.Unit;

/**
 * This class contains a single CCSDS packet.
 * The software usually reuses all Packet objects, so you should not save
 * a reference to a Packet in the expectations that the contents will not
 * change.
 * @author Stan Hilinski, January 2001
 */
public final class Packet extends Unit
{
    public static final int PRIMARY_HEADER_LENGTH = 6;
    public static final int TRANSPORT_HEADER_LENGTH = 10;
    public static final int LENGTH_OFFSET = 4;
    public static final int IDLE_PACKET = 0x07ff;
    private static final int ANNOTATION_HEADER_RECORD = 4;

    /**
     * This is packet annotation associated with this packet.
     */
    private Annotation annotation = new Annotation();

    /**
     * Create a packet.
     * @param length the packet length in bytes
     */
    public Packet(int length)
    {
        super(length);
    }

    /**
     * Reset this packet so that it may be used to contain a different packet.
     * @param length The length of the new packet
     */
    public void reset(int length)
    {
        deleted = false;
        super.length = length;
        annotation.reset();
        if (data.length < length)
        {
            data = new byte[length];
        }
    }

    /**
     * Get this packet's annotation text, if headers are present.
     */
    public String getAnnotationText() {
    	String s = null;
    	// first make sure there are headers
    	if (secondaryHeaderFlagSet()) {
    		// note total length of header section
    		int headersLength = totalHeaderLength();
    		System.out.println("Length of all headers: " + headersLength);
    		if (headersLength < 0) {
    			return s;
    		}
    		// loop through all headers present, or until we found a match
    		boolean moreHeaders = true;
    		int offset = TRANSPORT_HEADER_LENGTH + PRIMARY_HEADER_LENGTH;
    		byte headerCode;
    		int headerLength;
    		while (moreHeaders) {
    			// if we've gone through all headers present, exit
    			if ((offset + 1) > headersLength) break;
    			headerCode = data[offset];
    			headerLength = bytesToShort(data, offset + 1);
    			if (headerLength <= 0) {
    				System.out.println("Coudn't make sense of headers, skipping this image...");
    				break;
    			}
    			// System.out.println("Header length: " + headerLength + ", code: " + headerCode);
    			if (headerCode == ANNOTATION_HEADER_RECORD) {
    				s = new String(data, offset + 3, headerLength - 3);
    				break;
    			} else {
    				offset += headerLength;
    			}
    		}
    	}
    	return s;
    }
    
    /**
     * Get this packet's application ID.
     */
    public final int getApplicationId()
    {
        int x = ((int)data[0] << 8) | ((int)data[1] & 0x0ff);
        // XXX TJJ
        // System.out.println("Application Id: " + (x & 0x07ff));
        return x & 0x07ff;
    }
    
    /**
     * Get this packet's compression flag.
     */
    public final int getCompressionFlag(int offset)
    {
        return data[offset];
    }
    
    /**
     * Get this packet's data field length.
     */
    public final int getDataFieldLength()
    {
        return (bytesToShort(data, 4) + 1);
    }

    /**
     * Get this packet's first byte (used typically to read header types).
     */
    public final byte getFirstDataByte()
    {
        return data[TRANSPORT_HEADER_LENGTH + PRIMARY_HEADER_LENGTH];
    }

    /**
     * Get data byte at offset given
     */
    public final byte getDataByteAtOffset(int offset)
    {
        if ((offset - TRANSPORT_HEADER_LENGTH - PRIMARY_HEADER_LENGTH) <= data.length) {
    	    return data[TRANSPORT_HEADER_LENGTH + PRIMARY_HEADER_LENGTH + offset];
        } else {
        	return -1;
        }
    }
    
    /**
     * Get this packet's sequence counter.
     */
    public final int getSequenceCounter()
    {
        int x = ((int)data[2] << 8) | ((int)data[3] & 0x0ff);
        return x & 0x3fff;
    }

    /**
     * Get this packet's sequence flags field.
     */
    public final int getSequenceFlags()
    {
        int x = ((int) data[2] >> 6);
        return x & 0x03;
    }

    /**
     * Is secondary header flag set?
     */
    public final boolean secondaryHeaderFlagSet() 
    {
    	if ((data[0] & 0x08) != 0) {
    		return true;
    	} else {
    		return false;
    	}
    }  

    /**
     * Set this packet's total header length (used if extra headers added)
     */
    public final void setTotalHeaderLength(int newVal)
    {
    	data[TRANSPORT_HEADER_LENGTH + PRIMARY_HEADER_LENGTH + 4] = (byte) (newVal >>> 24);
    	data[TRANSPORT_HEADER_LENGTH + PRIMARY_HEADER_LENGTH + 5] = (byte) (newVal >>> 16);
    	data[TRANSPORT_HEADER_LENGTH + PRIMARY_HEADER_LENGTH + 6] = (byte) (newVal >>> 8);
    	data[TRANSPORT_HEADER_LENGTH + PRIMARY_HEADER_LENGTH + 7] = (byte) (newVal);
    }
    
    /**
     * Get this packet's total header length if it starts with a header
     */
    public final int totalHeaderLength()
    {
    	if (secondaryHeaderFlagSet()) {
    		return bytesToInt(data, TRANSPORT_HEADER_LENGTH + PRIMARY_HEADER_LENGTH + 4);
    	} else {
    		return -1;
    	}
    }
    
	/**
	 * Converts an (unsigned) byte to an unsigned int. 
	 * Since Java doesn't have an unsigned
	 * byte type, this requires some foolery.
	 * This solution based on information and code from
	 * http://www.rgagnon.com/javadetails/java-0026.html
	 * @param b The unsigned byte to convert
	 * @return the unsigned int equivalent
	 */
	public static int unsignedByteToInt(byte b) {
		return (int) b & 0xFF;
	}

	/**
	 * Converts a set of four consecutive bytes into a single int.
	 * @param data An array of bytes
	 * @param offset The array index to begin with
	 * @return The resulting int
	 */
	public static int bytesToInt(byte[] data, int offset) {
		int i = 0;
		i += unsignedByteToInt(data[offset]) << 24;
		i += unsignedByteToInt(data[offset + 1]) << 16;
		i += unsignedByteToInt(data[offset + 2]) << 8;
		i += unsignedByteToInt(data[offset + 3]);
		return i;
	}
	
	/**
	 * Converts a set of two consecutive bytes into a single int.
	 * @param data An array of bytes
	 * @param offset The array index to begin with
	 * @return The resulting int
	 */
	public static int bytesToShort(byte[] data, int offset) {
		int i = 0;
		i += unsignedByteToInt(data[offset]) << 8;
		i += unsignedByteToInt(data[offset + 1]);
		return i;
	}	
	
    /**
     * Get debug information.
     */
    public String toString()
    {
        return "length="+length+" deleted="+deleted+" "+annotation;
    }


    /**
     * Get this packet's annotation.
     */
    public final Packet.Annotation getPacketAnnotation()
    {
        return annotation;
    }

    /**
     * Packet annotation. It contains quality information for a single packet.
     */
    public final class Annotation
    {
        /**
         * If true, the packet has an invalid length. An operator will
         * define valid minimum and maximum lengths for each packet stream.
         */
        public boolean isInvalidLength = false;

        /**
         * If true, there was a packet sequence error between this one and
         * the preceding packet.
         */
        public boolean hasSequenceError = false;

        /**
         * If true, the packet could not be constructed in its entirety, so
         * the Path Service appended fill bytes to it to fill it out to its
         * advertised length.
         */
        public boolean isPacketWithFill = false;

        /**
         * The number of "good" bytes in this packet, which is the packet
         * length for most packets. If <code>isPacketWithFill</code> is
         * true, then this field is the index of the first fill byte.
         */
        public int goodByteCount;

        /**
         * Reset this annotation so that it can be used for a different
         * packet.
         */
        public void reset()
        {
            isPacketWithFill = false;
            goodByteCount = length;
            isInvalidLength = false;
            hasSequenceError = false;
        }

        public String toString()
        {
            return "badLen="+isInvalidLength +
                " goodByte="+goodByteCount +
                " withFill="+isPacketWithFill +
                " seqError="+hasSequenceError;
        }
    }
}