package nasa.stps.core.ccsds;
import nasa.stps.core.status.LongStatusItem;
import nasa.stps.core.*;

/**
 * This class does the CCSDS VCDU service. It makes VCDUs from CCSDS
 * version 2 frames.
 * @author Stan Hilinski, February 2001
 */
public final class VcduService extends AbstractService implements Sender,
        FrameReceiver, Cloneable
{
    /**
     * This is a class name for this STPS node type, which is also the element
     * name. It is not necessarily the link name, which is the name of one
     * particular object.
     */
    public static final String CLASSNAME = "vcdu";

    private Vcdu vcdu;
    private UnitReceiver output = null;
    private UnitOutputTool uotool = null;
    private LongStatusItem vcdus;


    /**
     * Create a VcduService object.
     */
    public VcduService()
    {
        super(CLASSNAME);
    }

    /**
     * Set up this stps node with a configuration.
     */
    public void load(org.w3c.dom.Element element, Configuration configuration)
            throws StpsException
    {
        super.load(element,configuration);
        vcdu = new Vcdu(element);
        vcdus = new LongStatusItem("VCDUs");
        statusItemList.add(vcdus);
    }

    /**
     * Add a Receiver to this sender's list of receivers.
     * @param node If the receiver is not of the expected type,
     *          then the method throws an StpsException.
     */
    public void addReceiver(Receiver receiver) throws StpsException
    {
        if (uotool == null)
        {
            uotool = new UnitOutputTool(getLinkName());
        }

        uotool.addReceiver(receiver);
        output = uotool.getOutput();
    }

    /**
     * Finish the setup. When this method is called, you may assume all nodes
     * have been created and exist by name in the map, and all standard links
     * have been resolved. This is a last chance to prepare for data flow.
     */
    public void finishSetup(Configuration configuration) throws StpsException
    {
        super.finishSetup(configuration);
        output = uotool.getOutput();
        uotool = null;   //no longer needed
        if (output == null)
        {
            throw new StpsException(getLinkName() + " demands an output link.");
        }
    }

    /**
     * Give a frame to this FrameReceiver.
     */
    public void putFrame(Frame frame) throws StpsException
    {
        if (!frame.isDeleted() && !frame.isFillFrame())
        {
            cadu.setFrame(frame);
            sequencer.check(cadu);
            vcdu.setCadu(cadu);
            output.putUnit(vcdu);
            ++vcdus.value;
        }
    }

    /**
     * Flush the pipeline.
     */
    public void flush() throws StpsException
    {
        output.flush();
    }

    /**
     * This class is a CCSDS VCDU.
     * I ignore (and eliminated) two fields in AbstractService because I don't
     * need them: OCFpresent and crcParityPresent.
     */
    class Vcdu extends Unit
    {
        private boolean discardRsParity = false;

        Vcdu(org.w3c.dom.Element element) throws StpsException
        {
            super();
            startOffset = 4;
            discardRsParity = Convert.toBoolean(element,"discardRsParity",false);
        }

        void setCadu(Cadu cadu)
        {
            frameAnnotation = cadu.getFrameAnnotation();
            data = cadu.getData();
            length = cadu.getSize() - startOffset;
            if (discardRsParity) length -= cadu.getReedSolomonParityLength();
            deleted = false;
        }
    }
}
