package nasa.stps.core.ccsds.path;
import nasa.stps.core.ccsds.Packet;
import java.util.ArrayList;

/**
 * This class manages a sequential list of packets to be filled.
 * It reuses its Packet objects, so subsequent users must not cache them.
 * @author Stan Hilinski, January 2001
 */
final class PacketList
{
    private ArrayList packetList = new ArrayList();
    private Packet currentPacket;
    private int currentPacketIndex = -1;
    private boolean isPartialCurrentPacket = true;

    /**
     * Get the next packet to fill from the packet list.
     * @packetLength The length of the desired packet.
     * @return A packet to fill
     */
    Packet get(int packetLength)
    {
        ++currentPacketIndex;
        if (packetList.size() == currentPacketIndex)
        {
            currentPacket = new Packet(packetLength);
            packetList.add(currentPacket);
        }
        else
        {
            currentPacket = (Packet)packetList.get(currentPacketIndex);
            currentPacket.reset(packetLength);
        }

        isPartialCurrentPacket = true;

        return currentPacket;
    }

    /**
     * Mark that the current packet, which was last obtained via get(), as a
     * completed one.
     */
    final void setCurrentPacketIsComplete()
    {
        isPartialCurrentPacket = false;
    }

    /**
     * Reset the packet list so it contains no full or partial packets.
     */
    final void flushAllData()
    {
        currentPacketIndex = -1;
        currentPacket = null;
    }

    /**
     * Remove all completed packets from the list but keep the last partial
     * one if it exists.
     */
    void flushCompletedPackets()
    {
        if (currentPacket == null) return;

        if (!isPartialCurrentPacket)
        {
            flushAllData();
        }
        else if (currentPacketIndex > 0)
        {
            /**
             * When the index > 0 and the last packet is partial, it means
             * I have a part-filled packet following a bunch of filled ones.
             * I want to put the partial one first and to forget about the
             * rest. I swap the first packet and part-filled packet positions.
             */
            Object x = packetList.get(0);
            packetList.set(0,currentPacket);
            packetList.set(currentPacketIndex,x);
            currentPacketIndex = 0;
        }
        /**
         * The remaining case is a part-filled packet in the first position,
         * which is what I want, so I do nothing.
         */
    }

    /**
     * Get a list of completed packets in time order.
     * @return null if there are no completed packets in the list.
     */
    Packet[] getList()
    {
        Packet[] list = null;
        int completedPackets = currentPacketIndex;

        /**
         * I don't advance the packet index until I need a new packet, so
         * I must count the current packet if it is full.
         */
        if (!isPartialCurrentPacket) ++completedPackets;

        if (completedPackets > 0)
        {
            list = new Packet[completedPackets];
            for (int n = 0; n < completedPackets; n++)
            {
                list[n] = (Packet)packetList.get(n);
            }
        }
        return list;
    }
}