package nasa.stps.core.xstps.hrit;

import nasa.stps.core.ccsds.Packet;
import nasa.stps.core.Configuration;
import nasa.stps.core.StpsException;
import nasa.stps.core.Builder;
import nasa.stps.core.Convert;
import nasa.stps.core.status.LongStatusItem;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;


/**
 * This STPS node is a java version of an HRIT decoder program. It creates 
 * HRIT image data segment files.
 * @author Tommy Jasmin, April 2006
 */

public final class HRITOutput extends nasa.stps.core.StpsNode
        implements nasa.stps.core.ccsds.PacketReceiver
{
    /**
     * This is the class name for this STPS node type, which is also the
     * element name. It is not necessarily the link name, which is the name
     * of one particular object.
     */
    public static final String CLASSNAME = "hrit";

    private Appid[] apList;
    private int totalPackets = 0;
    private long totalBytes = 0L;
    private OutputFiles output;
    private int spid = -1;
    private boolean test = false;
    private LongStatusItem packetsWritten;
    private Builder builder;

    /**
     * A null constructor.
     */
    public HRITOutput(Builder b)
    {
        super(CLASSNAME);
        builder = b;
    }

    /**
     * Configure from an XML document. You cannot assume that any other stps
     * nodes have been created.
     */
    public void load(org.w3c.dom.Element element,
            Configuration configuration) throws StpsException
    {
        statusItemList = new java.util.ArrayList(3);
        packetsWritten = new LongStatusItem("Packets Written");
        statusItemList.add(packetsWritten);

        String name = element.getAttribute("label");
        super.setLinkName(name);

        spid = Convert.toInteger(element,"spid",spid);
        test = Convert.toBoolean(element,"test",test);
        
        // get the spacecraft name from the main setup document
        Document doc = element.getOwnerDocument();
        NodeList nl = doc.getElementsByTagName("spacecraft");
        Element sce = (Element) nl.item(0);
        String spacecraftName = sce.getAttribute("label");

        org.w3c.dom.NodeList nodes = element.getElementsByTagName("appid");
        int appidCount = nodes.getLength();

        if (appidCount < 1 || appidCount > 3)
        {
            throw new StpsException("hrit: There must be 1-3 appid elements specified.");
        }

        apList = new Appid[appidCount];

        for (int n = 0; n < appidCount; n++)
        {
            org.w3c.dom.Element e = (org.w3c.dom.Element) nodes.item(n);
            apList[n] = new Appid(e, spid, false);
        }

        String path = element.getAttribute("path");

        try
        {
            output = new OutputFiles(path);
            output.setSpacecraftName(spacecraftName);
            output.setBuilder(builder);
            for (int n = 0; n < appidCount; n++)
            {
                output.setAppidSpid(n,apList[n].getId(),apList[n].getSpacecraftId());
            }
        }
        catch (java.io.IOException ioe)
        {
            throw new StpsException(ioe);
        }
    }

    /**
     * Finish the setup. When this method is called, you may assume all nodes
     * have been created and exist by name in the map, and all standard links
     * have been resolved. This is a last chance to prepare for data flow.
     */
    public void finishSetup(Configuration configuration) throws StpsException
    {
    }

    /**
     * Give an array of packets to this PacketReceiver.
     */
    public void putPackets(Packet[] packets) throws StpsException
    {
        for (int n = 0; n < packets.length; n++)
        {
            putPacket(packets[n]);
        }
    }

    /**
     * Give a packet to this PacketReceiver.
     */
    public void putPacket(Packet packet) throws StpsException
    {
        ++packetsWritten.value;

        //First I identify to which appid the packet belongs.
        Appid ap = null;
        int apindex = 0;
        int a = packet.getApplicationId();
        for (int n = 0; n < apList.length; n++)
        {
            if (apList[n].getId() == a)
            {
                ap = apList[n];
                apindex = n;
                break;
            }
        }

        //If ap==null, then I don't recognize the appid.
        if (ap != null)
        {
            //I store packet information in its appid. I load "current"
            //with "packet" in ap and not here because ap knows critical
            //information about how the packet is formed.
            boolean save = ap.putPacket(packet,totalBytes);

            if (save)
            {
                ++totalPackets;
                totalBytes += packet.getSize();

                try
                {
                    output.write(packet,apindex);
                }
                catch (java.io.IOException ioe)
                {
                    throw new StpsException(ioe);
                }
            }
        }
    }

    /**
     * The session is over. Create the construction record file.
     */
    public void flush() 
    {
    	// do nothing for HRIT
    	return;
    }
    
}
