package nasa.stps.core.xstps.hrit;

import java.io.IOException;
import java.io.InputStream;

import javax.imageio.stream.FileImageOutputStream;

/**
 * @author tommyj
 * The Huffman Decoder.  Adapted from C-based versions developed at 
 * Australian Bureau of Meteorology and Cornell University.
 */

public class HuffD {
		
	int getBuffer = 0;
	int bitsLeft = 0;
	public static final int BITS_PER_INT = (8 * 4);
	public static final int MIN_GET_BITS = (BITS_PER_INT - 7);
	private InputStream is = null;
	private FileImageOutputStream fios = null;
	boolean debug = false;
	boolean nonTable = true;
	int globalRow = -1;
	int globalCol = -1;
	int printedByte = 0;
	int resetChar1 = 0x00;
	int resetChar2 = 0x00;
	boolean unGetChar1 = false;
	boolean unGetChar2 = false;
	// local copy of Huffman Table numbits
	int [] numbits = new int[256];
	// local copy of Huffman Table value
	int [] valueArray = new int[256];	
	// local copy of JPEG component info object
	JpegComponentInfo jci = null;
	// local copy of Huffman Table
	HuffmanTable ht = null;
	// buffer to output one row at a time, single-byte data
	byte [] outputBufferOneByte = null;
	// buffer to output one row at a time, two-byte data
	byte [] outputBufferTwoByte = null;
	// local copy image width
	int imageWidth = -1;
	// local copy image height
	int imageHeight = -1;
	
	public static final int bitMask[] = {  0xffffffff, 0x7fffffff, 0x3fffffff, 0x1fffffff,
            0x0fffffff, 0x07ffffff, 0x03ffffff, 0x01ffffff,
            0x00ffffff, 0x007fffff, 0x003fffff, 0x001fffff,
            0x000fffff, 0x0007ffff, 0x0003ffff, 0x0001ffff,
            0x0000ffff, 0x00007fff, 0x00003fff, 0x00001fff,
            0x00000fff, 0x000007ff, 0x000003ff, 0x000001ff,
            0x000000ff, 0x0000007f, 0x0000003f, 0x0000001f,
            0x0000000f, 0x00000007, 0x00000003, 0x00000001};
	
	/*
	 * bmask[n] is mask for n rightmost bits
	 */
	public static final int bmask[] = {0x00000000,
	         0x00000001, 0x00000003, 0x00000007, 0x0000000F,
	         0x0000001F, 0x0000003F, 0x0000007F, 0x000000FF,
	         0x000001FF, 0x000003FF, 0x000007FF, 0x00000FFF,
	         0x00001FFF, 0x00003FFF, 0x00007FFF, 0x0000FFFF};
	
	private static int extendTest[] =     /* entry n is 2**(n-1) */
	{0, 0x0001, 0x0002, 0x0004, 0x0008, 0x0010, 0x0020, 0x0040, 0x0080,
	 0x0100, 0x0200, 0x0400, 0x0800, 0x1000, 0x2000, 0x4000, 0x8000};

	private static int extendOffset[] =   /* entry n is (-1 << n) + 1 */
	{0, ((-1) << 1) + 1, ((-1) << 2) + 1, ((-1) << 3) + 1, ((-1) << 4) + 1,
	 ((-1) << 5) + 1, ((-1) << 6) + 1, ((-1) << 7) + 1, ((-1) << 8) + 1,
	 ((-1) << 9) + 1, ((-1) << 10) + 1, ((-1) << 11) + 1, ((-1) << 12) + 1,
	 ((-1) << 13) + 1, ((-1) << 14) + 1, ((-1) << 15) + 1, ((-1) << 16) + 1};
	
	/**
	 * @param fios
	 */
	public HuffD(InputStream is, FileImageOutputStream fios) {
		super();
		this.is = is;
		this.fios = fios;
	}

	/*
	 *--------------------------------------------------------------
	 *
	 * HuffDecoderInit --
	 *
	 *      Initialize for a Huffman-compressed scan.
	 *      This is invoked after reading the SOS marker.
	 *
	 * Results:
	 *      None
	 *
	 * Side effects:
	 *      None.
	 *
	 *--------------------------------------------------------------
	 */
	
	int HuffDecoderInit(DecompressInfo dcPtr)
	{

		/*
		 * Initialize static variables
		 */
		bitsLeft = 0;
		jci = dcPtr.getCompInfo();
		ht = dcPtr.getDcHuffTbl();
		imageHeight = dcPtr.getImageHeight();
		imageWidth = dcPtr.getImageWidth();
		outputBufferOneByte = new byte[imageWidth];
		outputBufferTwoByte = new byte[imageWidth * 2];
		
		/*
		 * Make sure requested tables are present
		 */
		if (ht == null) {
			return -1;
		}

		/*
		 * Compute derived values for Huffman tables.
		 * We may do this more than once for same table, but it's not a
		 * big deal
		 */
		FixHuffTbl(ht);

		/*
		 * Initialize restart stuff
		 */
		dcPtr.setRestartInRows(dcPtr.getRestartInterval() / imageWidth);
		dcPtr.setRestartRowsToGo(dcPtr.getRestartInRows());
		dcPtr.setNextRestartNum(0);

		return 0;
	}

	/*
	 *--------------------------------------------------------------
	 *
	 * FixHuffTbl --
	 *
	 *      Compute derived values for a Huffman table one the DHT marker
	 *      has been processed.  This generates both the encoding and
	 *      decoding tables.
	 *
	 * Results:
	 *      None.
	 *
	 * Side effects:
	 *      None.
	 *
	 *--------------------------------------------------------------
	 */
	
	private void FixHuffTbl(HuffmanTable htbl)
	{
	   int p, i, l, lastp, si;
	   byte [] huffsize = new byte[257];
	   int [] huffcode = new int[257];
	   int code;
	   int size;
	   int value, ll, ul;

	   /*
	    * Figure C.1: make table of Huffman code length for each symbol
	    * Note that this is in code-length order.
	    */
	   p = 0;
	   for (l = 1; l <= 16; l++) {
	      for (i = 1; i <= (int) (htbl.getBits()[l] & 0xFF); i++) {
	         huffsize[p++] = (byte) l;
	      }
	   }
	   huffsize[p] = 0;
	   lastp = p;

	   /*
	    * Figure C.2: generate the codes themselves
	    * Note that this is in code-length order.
	    */
	   code = 0;
	   si = (int) (huffsize[0] & 0xFF);
	   p = 0;
	   while (huffsize[p] != 0) {
	      while (((int) (huffsize[p] & 0xFF)) == si) {
	         huffcode[p++] = code;
	         code++;
	      }
	      code <<= 1;
	      si++;
	   }

	   /*
	    * Figure F.15: generate decoding tables
	    */
	   p = 0;
	   for (l = 1; l <= 16; l++) {
	      if (htbl.getBits()[l] != 0) {
	         htbl.getValptr()[l] = p;
	         htbl.getMincode()[l] = huffcode[p];
	         p = p + htbl.getBits()[l];
	         htbl.getMaxcode()[l] = huffcode[p - 1];
	      }
	      else {
	         htbl.getMaxcode()[l] = -1;
	      }
	   }

	   /*
	    * We put in this value to ensure HuffDecode terminates.
	    */
	   htbl.getMaxcode()[17] = 0x000FFFFF;

	   /*
	    * Build the numbits, value lookup tables.
	    * These table allow us to gather 8 bits from the bits stream,
	    * and immediately lookup the size and value of the huffman codes.
	    * If size is zero, it means that more than 8 bits are in the huffman
	    * code (this happens about 3-4% of the time).
	    */

	   htbl.clearNumbits();
	   for (p = 0; p < lastp; p++) {
	      size = (int) (huffsize[p] & 0xFF);
	      if (size <= 8) {
	         value = htbl.getHuffval()[p];
	         code = huffcode[p];
	         ll = code << (8 - size);
	         if (size < 8) {
	            ul = ll | bitMask[24 + size];
	         }
	         else {
	            ul = ll;
	         }
	         for (i = ll; i <= ul; i++) {
	            htbl.getNumbits()[i] = size;
	            htbl.getValue()[i] = value;
	         }
	      }
	   }
	   
	   // set local array refs
	   numbits = htbl.getNumbits();
	   valueArray = htbl.getValue();
	   if (debug) {
		   System.err.println("Huffman table: \n" + htbl);
	   }

	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * DecodeImage --
	 *
	 *      Decode the input stream. This includes modifying
	 *      the component value so the real value, not the
	 *      difference is returned.
	 *
	 * Results:
	 *      None.
	 *
	 * Side effects:
	 *      Bitstream is parsed.
	 *
	 *--------------------------------------------------------------
	 */
	
	int DecodeImage(DecompressInfo dcPtr, int [] mcuROW1, int [] mcuROW2)
	{
	   int s;
	   int d = 0;
	   int col, row;
	   int predictor;
	   int numCOL, numROW, Pr;
	   int[] prevRowBuf;
	   int[] curRowBuf;
	   int Pt, psv;

	   numCOL = imageWidth;
	   numROW = imageHeight;
	   Pt = dcPtr.getPt();
	   psv = dcPtr.getSs();
	   prevRowBuf = mcuROW2;
	   curRowBuf = mcuROW1;
	   Pr = dcPtr.getDataPrecision();

	   /*
	    * Decode the first row of image. Output the row and
	    * turn this row into a previous row for later predictor
	    * calculation.
	    */

	   globalRow = 0;
	   DecodeFirstRow(dcPtr, curRowBuf);
	   PmPutRow(curRowBuf, numCOL, Pt, Pr);
	   // swap prevRowBuf, curRowBuf
	   int [] tmpBuf = prevRowBuf;
	   prevRowBuf = curRowBuf;
	   curRowBuf = tmpBuf;
	   tmpBuf = null;

	   for (row = 1; row < numROW; row++) {

	      /*
	       * Account for restart interval, process restart marker if needed.
	       */
		   
	      if (dcPtr.getRestartInRows() != 0) {
	         if (dcPtr.getRestartRowsToGo() == 0) {
	        	System.err.println("ProcessRestart called at row " + row);
	            ProcessRestart (dcPtr);

	            /*
	             * Reset predictors at restart.
	             */
	            DecodeFirstRow(dcPtr, curRowBuf);
	            PmPutRow(curRowBuf, numCOL, Pt, Pr);
	            // swap prevRowBuf, curRowBuf
	     	    int [] tmpBuf2 = prevRowBuf;
	    	    prevRowBuf = curRowBuf;
	    	    curRowBuf = tmpBuf2;
	    	    tmpBuf2 = null;
	            continue;
	         }

	         dcPtr.setRestartRowsToGo(dcPtr.getRestartRowsToGo() - 1);
	      }

	      /*
	       * The upper neighbors are predictors for the first column.
	       */

	      //dctbl = dcPtr.getDcHuffTblPtrs()[0];

	      /*
	       * Section F.2.2.1: decode the difference
	       */
	      s = HuffDecode(ht);
	      if (s > 0) {
	    	  if (s < 16) {
	    		  d = getBits(s);
	    		  if (d < extendTest[s]) d += extendOffset[s];
	    	  }
	    	  else {;
	    	  	  d = 32768;
	    	  }
	      }
	      else {
	    	  d = 0;
	      }

	      curRowBuf[0] = d + prevRowBuf[0];


	      /*
	       * For the rest of the column on this row, predictor
	       * calculations are base on PSV.
	       */
	      for (col = 1; col < numCOL; col++) {

	    	  //dctbl = dcPtr.getDcHuffTblPtrs()[0];

	    	  /*
	    	   * Section F.2.2.1: decode the difference
	    	   */
	    	  s = HuffDecode(ht);
	    	  if (s > 0) {
	    		  if (s < 16) {
	    			  d = getBits(s);
	    			  if (d < extendTest[s]) d += extendOffset[s];                  
	    		  }
	    		  else {
	    			  d = 32768;
	    		  }
	    	  }
	    	  else {
	    		  d = 0;
	    	  }

	    	  predictor = QuickPredict(col, curRowBuf, prevRowBuf, psv);
	    	  curRowBuf[col] = d + predictor;

	      }

	      PmPutRow(curRowBuf, numCOL, Pt, Pr);
	      // swap prevRowBuf, curRowBuf
	      int [] tmpBuf3 = prevRowBuf;
	      prevRowBuf = curRowBuf;
	      curRowBuf = tmpBuf3;
	      tmpBuf3 = null;
	   }

	   return 0;
	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * DecodeFirstRow --
	 *
	 *      Decode the first raster line of samples at the start of
	 *      the scan and at the beginning of each restart interval.
	 *      This includes modifying the component value so the real
	 *      value, not the difference is returned.
	 *
	 * Results:
	 *      None.
	 *
	 * Side effects:
	 *      Bitstream is parsed.
	 *
	 *--------------------------------------------------------------
	 */
	
	void DecodeFirstRow(DecompressInfo dcPtr, int [] curRowBuf)
	{
	   int s = 0;
	   int col, numCOL;
	   int Pr, Pt;
	   int d = 0;

	   Pr = dcPtr.getDataPrecision();
	   Pt = dcPtr.getPt();
	   numCOL = imageWidth;

	   /*
	    * the start of the scan or at the beginning of restart interval.
	    */

	   //dctbl = dcPtr.getDcHuffTblPtrs()[0];

	   /*
	    * Section F.2.2.1: decode the difference
	    */
	   s = HuffDecode(ht);
	   if (s > 0) {
		   if (s < 16) {
			   d = getBits(s);
			   if (d < extendTest[s]) d += extendOffset[s];
		   } else {
			   d = 32768;
		   }
	   }
	   else {
		   d = 0;
	   }

	   /*
	    * Add the predictor to the difference.
	    */
	   curRowBuf[0] = d + (1 << (Pr - Pt - 1));

	   /*
	    * the rest of the first row
	    */
	   for (col = 1; col < numCOL; col++) {

		   //dctbl = dcPtr.getDcHuffTblPtrs()[0];

		   /*
		    * Section F.2.2.1: decode the difference
		    */
		   s = HuffDecode(ht);
		   if (s > 0) {
			   if (s < 16) {
				   d = getBits(s);
				   //d = HuffExtend2(d, s);
				   if (d < (1 << (s - 1))) {            
					   d = d + ((-1 << s) + 1);    
				   }   
			   }
			   else {
				   d = 32768;
			   }
		   }
		   else {
			   d = 0;
		   }

		   /*
		    * Add the predictor to the difference.
		    */
		    curRowBuf[col] = d + curRowBuf[col - 1];

	   }

	   if (dcPtr.getRestartInRows() != 0) {
	      dcPtr.setRestartRowsToGo(dcPtr.getRestartRowsToGo() - 1);
	   }
	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * HuffExtend --
	 *
	 *      Code and table for Figure F.12: extend sign bit
	 *
	 * Results:
	 *      The extended value.
	 *
	 * Side effects:
	 *      None.
	 *
	 *--------------------------------------------------------------
	 */

	int HuffExtend(int x, int s) {         
	   if (x < extendTest[s]) {       
	      x += extendOffset[s];       
	   }   
	   return x;
	}
	
/*	int HuffExtend2(int x, int s)  {            
		if (x < (1 << (s - 1))) {            
			x = x + ((-1 << s) + 1);    
		}      
		return x;
	}*/
	
	/*
	 *--------------------------------------------------------------
	 *
	 * ProcessRestart --
	 *
	 *      Check for a restart marker & resynchronize decoder.
	 *
	 * Results:
	 *      None.
	 *
	 * Side effects:
	 *      BitStream is parsed, bit buffer is reset, etc.
	 *
	 *--------------------------------------------------------------
	 */
	int ProcessRestart(DecompressInfo dcPtr)
	{
		int c;

		/*
		 * Throw away any unused bits remaining in bit buffer
		 */

		bitsLeft = 0;

		/*
		 * Scan for next JPEG marker
		 */
		do {
			do {                 /* skip any non-FF bytes */
				c = GetJpegChar();
			} while (c != 0xFF);

			do {                      /* skip any duplicate FFs */
				/*
				 * we don't increment nbytes here since extra FFs are legal
				 */
				c = GetJpegChar();
			} while (c == 0xFF);

		} while (c == 0);            /* repeat if it was a stuffed FF/00 */

		if (c != (JpegMarker.M_RST0 + dcPtr.getNextRestartNum())) {
			/*
			 * Uh-oh, the restart markers have been messed up too.
			 * Just bail out.
			 */

			return -1;
		}

		/*
		 * Update restart state
		 */
		dcPtr.setRestartRowsToGo(dcPtr.getRestartInRows());
		dcPtr.setNextRestartNum((dcPtr.getNextRestartNum() + 1) & 7);

		return 0;
	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * QuickPredict --
	 *
	 *      Calculate the predictor for sample curRowBuf[col].
	 *      It does not handle the special cases at image edges, such
	 *      as first row and first column of a scan. We put the special
	 *      case checkings outside so that the computations in main
	 *      loop can be simpler. This has enhanced the performance
	 *      significantly.
	 *
	 * Results:
	 *      predictor is returned.
	 *
	 * Side effects:
	 *      None.
	 *
	 *--------------------------------------------------------------
	 */	
	
	int QuickPredict(int col, int [] curRowBuf, int [] prevRowBuf, int psv) { 
		int left, upper, diag, leftcol;                              

		leftcol = col - 1;                                                      
		upper = prevRowBuf[col];                                    
		left = curRowBuf[leftcol];                                 
		diag = prevRowBuf[leftcol];     
		int predictor = 0;

		/*                                                                 
		 * All predictor are calculated according to psv.                  
		 */                                                                 
		switch (psv) {                                                     
		case 0:                                                          
			predictor = 0;                                          
			break;                                                    
		case 1:                                                           
			predictor = left;                                        
			break;                                                    
		case 2:                                                           
			predictor = upper;                                       
			break;                                                    
		case 3:                                                           
			predictor = diag;                                        
			break;                                                   
		case 4:                                                           
			predictor = left + upper - diag;                            
			break;                                                    
		case 5:                                                          
			predictor = left + ((upper - diag) >>> 1);                      
			break;                                                    
		case 6:                                                           
			predictor = upper + ((left - diag) >>> 1);                     
			break;                                                  
		case 7:                                                          
			predictor = (left + upper) >>> 1;                            
			break;                                                   
		default:                                                         
			System.err.println("Warning: Undefined PSV");               
			predictor = 0;  
			break;
		}  
		return predictor;
	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * HuffDecode --
	 *
	 *      Taken from Figure F.16: extract next coded symbol from
	 *      input stream.  This should become a macro.
	 *
	 * Results:
	 *      Next coded symbol
	 *
	 * Side effects:
	 *      Bitstream is parsed.
	 *
	 *--------------------------------------------------------------
	 */	
	
	int HuffDecode(HuffmanTable htbl)                                                                     
	{                                                                                                                
		int l = 0;
		int code = 0;
		int temp = 0;   
		int rv = 0;		

		/*                                                                                                          
		 * If the huffman code is less than 8 bits, we can use the fast      
		 * table lookup to get its value.  It's more than 8 bits about         
		 * 3-4% of the time.                                                                      
		 */    
		
		code = showBits8();   
		if (numbits[code] != 0) {                                                                   
			bitsLeft -= (numbits[code]);                                                  
			rv = valueArray[code];                                                                       
		}                                                                 
		else {      
			bitsLeft -= 8;                                                                                     
			l = 8;                                                                                                
			while (code > htbl.getMaxcode()[l]) {                                              
				temp = getBit();                                                                       
				code = (code << 1) | temp;    
				l++;                                                                                          
			}                                                                                                   

			/*                                                                                                       
			 * With garbage input we may reach the sentinel value l = 17.        
			 */                                                                                              

			if (l > 16) {  
				if (debug) {
					System.err.println("Corrupt JPEG data: bad Huffman code");
				}  
				rv = 0;           /* fake a zero as the safest result */           
			} else {                                                                                        
				rv = htbl.getHuffval()[htbl.getValptr()[l] + code - htbl.getMincode()[l]];                                     
			}                                                                                                   
		}    
		return rv;
	}
	
	int showBits8() {     
		if (bitsLeft < 8) FillBitBuffer(8);                             
		return (int) ((getBuffer >>> (bitsLeft - 8)) & 0xff);  
	}

	int getBit() {     
		if (bitsLeft == 0) FillBitBuffer(1);                                
		return (int) ((getBuffer >>> (--bitsLeft)) & 0x01);   
	}
	
	int getBits(int nbits) {    
		if (bitsLeft < nbits) FillBitBuffer(nbits);                     
		return (int) ((getBuffer >>> (bitsLeft -= (nbits)))) & bmask[nbits];   
	}

	/*
	 *--------------------------------------------------------------
	 *
	 * FillBitBuffer --
	 *
	 *      Load up the bit buffer with at least nbits
	 *      Process any stuffed bytes at this time.
	 *
	 * Results:
	 *      None
	 *
	 * Side effects:
	 *      The bitwise global variables are updated.
	 *
	 *--------------------------------------------------------------
	 */

	void FillBitBuffer(int nbits)
	{
		byte b1;
		byte b2;

		while (bitsLeft < MIN_GET_BITS) {

			b1 = (byte) GetJpegChar();

			/*
			 * If it's 0xFF, check and discard stuffed zero byte
			 */
			if (b1 == -1) {

				//System.err.println("b1 was 0xFF, stuffed zero...");
				b2 = (byte) GetJpegChar();

				if (b2 != 0) {

					/*
					 * Oops, it's actually a marker indicating end of
					 * compressed data.  Better put it back for use later.
					 */
					//System.err.println("UnGetJpegChar called...");
					resetChar2 = (b2 & 0xFF);
					resetChar1 = (b1 & 0xFF);
					unGetChar2 = true;
					unGetChar1 = true;
					
					// UnGetJpegChar();

					/*
					 * There should be enough bits still left in the data
					 * segment; if so, just break out of the while loop.
					 */
					if (bitsLeft >= nbits) break;

					/*
					 * Uh-oh.  Corrupted data: stuff zeroes into the data
					 * stream, since this sometimes occurs when we are on the
					 * last show_bits(8) during decoding of the Huffman
					 * segment.
					 */
					b1 = 0;
				}
			}

			/*
			 * OK, load c into getBuffer
			 */
			getBuffer = (getBuffer << 8) | (b1 & 0xff);
			bitsLeft = bitsLeft + 8;
		}
	}

	/*
	 *--------------------------------------------------------------
	 *
	 * PmPutRow --
	 *
	 *      Output one row of pixels stored in RowBuf.
	 *
	 * Results:
	 *      None
	 *
	 * Side effects:
	 *      One row of pixels are write to file pointed by outFile.
	 *
	 *--------------------------------------------------------------
	 */

	void PmPutRow(int [] RowBuf, int numCol, int Pt, int Pr)
	{
		int col;
		int vLo, vHi;

		/*
		 * Multiply 2^Pt before output. Pt is the point
		 * transform parameter.
		 */
		try {

			for (col = 0; col < numCol; col++) {
				vLo = RowBuf[col] << Pt;
				if (Pr > 8) {
					vHi = vLo >> 8;
					//fios.write(vHi & 0xFF);
					outputBufferTwoByte[col * 2]     = (byte) (vHi & 0xFF);
					outputBufferTwoByte[col * 2 + 1] = (byte) (vLo & 0xFF);
				} else {
					//fios.write(vLo & 0xFF);
					outputBufferOneByte[col] = (byte) (vLo & 0xFF);
				}
			}
			if (Pr > 8) {
				fios.write(outputBufferTwoByte, 0, outputBufferTwoByte.length);
			} else {
				fios.write(outputBufferOneByte, 0, outputBufferOneByte.length);
			}

		} catch (IOException ioe) {
			System.err.println("Error: problem decoding JPEG data, details follow: ");
			ioe.printStackTrace();       		
		}
	}
	
	public int GetJpegChar() {
		int c = 0;
		
		// if stream was reset, we are holding an 0xFF to return
		if (unGetChar1) {
			//System.err.println("Stream reset!!, returning: " + resetChar);
			unGetChar1 = false;
			return resetChar1;
		}
		
		if (unGetChar2) {
			//System.err.println("Stream reset!!, returning: " + resetChar);
			unGetChar2 = false;
			return resetChar2;
		}
		
		try {
			// the count 100 is completely arbitrary - not used
			//bis.mark(100);
			c = is.read();
		} catch (IOException ioe) {
			// TJJ XXX how to handle this?
			System.err.println("ERROR: Unexpected EOF!");
		}
		
		return (int) c & 0xFF;
	}


/*	private void UnGetJpegChar() {
		try {
			bis.reset();
			streamReset = true;
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}*/
	
}
