package nasa.stps.core.xstps.hrit;

import java.io.IOException;
import java.io.InputStream;

import javax.imageio.stream.FileImageOutputStream;

/**
 * @author tommyj
 * Lossless JPEG decoder parent class.  Finds and interprets header info,
 * then cranks up a Huffman Decoder.
 */

public class LosslessJpeg {
	
	public static final int MinPrecisionBits = 2;
	public static final int MaxPrecisionBits = 16;
	public static final int MinPrecisionValue = 2;
	public static final int MaxPrecisionValue = 65535;
	
	/*
	 * Global input and output file pointers.
	 */
	InputStream inFile = null;                             /**< File pointer to a file opened for reading. */                           /**< File pointer to a file opened for writing. */

	int numInputBytes;                       /**< The total number of bytes in inputBuffer */
	int inputBufferOffset;                   /**< Offset of current byte */	
	
	// current and previous scan buffers
	int [] mcuROW1 = null;
	int [] mcuROW2 = null;
	
	HuffD huffd = null;
	
	/**
	 * Entry point of the lossless jpeg decompressor function.
	 *
	 * It is passed the following parameters.
	 * @param verbose - an integer flag to control verbose program output.
	 *                - 0 for quiet mode.
	 *                - 1 for verbose mode.
	 * @param inFlag - an integer that specifies the input file format.
	 *               - 0 means the input file is a raw lossless jpeg file.
	 *               - 1 means it is a MTSAT .TRN file.
	 * @param outFlag - an integer that determines the output file format.
	 *                - 0 means just output the raw pixels.
	 *                - 1 means decompressed pnm file.
	 *                - 2 means output a decompressed trn file.
	 * @param innyFile - a pointer to a FILE pointer designating the input file \a inFile.
	 * @param outtyFile - a pointer to a FILE pointer designating the output file.
	 *
	 * @return an integer in the following ranges:
	 *    - 0:    for a successful decompress.
	 *    - -1 to -9: if reading the transport file header failed.
	 *    - -21
	 *
	 */
	public int ljpeg(InputStream is, FileImageOutputStream fios)
	{
	   int retVal;
	   DecompressInfo dcInfo = new DecompressInfo();
	   huffd = new HuffD(is, fios);

	   /*
	    * Grab the passed in file pointers and copy them into global variables.
	    */
	   inFile = is;

	   /*
	    * Initialize the input data buffer pointer and counter variables.
	    */
	   numInputBytes=0;               /* The total number of bytes in inputBuffer */
	   inputBufferOffset=0;           /* Offset of current byte */

	   /*
	    * Read the JPEG File header, up to scan header, and initialize all
	    * the variables in the decompression information structure.
	    */
	   retVal = ReadFileHeader(dcInfo);
	   if (retVal < 0) {
	      return (retVal-30);
	   }

	   /*
	    * Loop through each scan in image. ReadScanHeader returns
	    * 0 once it consumes and EOI marker.
	    */
	   if (! ReadScanHeader(dcInfo)) {
	      if (true) System.err.println("Empty JPEG file\n");
	      return -1;
	   }

	   /*
	    * Output image parameter if verbose flag is on.
	    */
	   if (true) {
	      System.err.println("** JPEG image parameters **");
	      System.err.println("   Sample precision = " + dcInfo.getDataPrecision());
	      System.err.println("   Image width = " + dcInfo.getImageWidth());
	      System.err.println("   Image height = " + dcInfo.getImageHeight());
	   }

	   /*
	    * Set up and initialize the relevant structures, decompress
	    * the jpeg bitstream, and then clean up.
	    */
	   
	   retVal = DecoderStructInit(dcInfo);
	   if (retVal < 0) {
		   mcuROW1 = null;
		   mcuROW2 = null;
		   return retVal - 50;
	   }

	   retVal = huffd.HuffDecoderInit(dcInfo);
	   if (retVal < 0) {
		   mcuROW1 = null;
		   mcuROW2 = null;
		   return retVal - 60;
	   }

	   retVal = huffd.DecodeImage(dcInfo, mcuROW1, mcuROW2);
	   if (retVal < 0) {
		   mcuROW1 = null;
		   mcuROW2 = null;
		   return retVal - 70;
	   }

	   /*
	    * Free memory used in the decoder.
	    */
	   mcuROW1 = null;
	   mcuROW2 = null;

	   /*
	    * All done!
	    */
	   return 0;

	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * DecoderStructInit --
	 *
	 *      Initialize the rest of the fields in the decompression
	 *      structure.
	 *
	 * Results:
	 *      None.
	 *
	 * Side effects:
	 *      None.
	 *
	 *--------------------------------------------------------------
	 */
	
	private int DecoderStructInit(DecompressInfo dcPtr) {

		JpegComponentInfo jci;

		/*
		 * Check sampling factor validity, should both be 1 acc to spec
		 */

		jci = dcPtr.getCompInfo();
		if ((jci.getHSampFactor() != 1) || (jci.getVSampFactor() != 1)) {
			return -1;
		}

		/*
		 * Prepare array describing MCU composition
		 */
		dcPtr.getMCUmembership()[0] = 0;

		
		/*
		 * Initialize mucROW1 and mcuROW2 which buffer two rows of
		 * pixels for predictor calculation.
		 */
		
		mcuROW1 = new int[dcPtr.getImageWidth()];
		mcuROW2 = new int[dcPtr.getImageWidth()];

		return 0;
	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * ReadScanHeader --
	 *
	 *      Read the start of a scan (everything through the SOS marker).
	 *
	 * Results:
	 *      1 if find SOS, 0 if find EOI
	 *
	 * Side effects:
	 *      Bitstream is parsed, may exit on errors.
	 *
	 *--------------------------------------------------------------
	 */

	private boolean ReadScanHeader(DecompressInfo dcInfo) {
		   int c;

		   /*
		    * Process markers until SOS or EOI
		    */
		   c = ProcessTables(dcInfo);

		   switch (c) {
		      case JpegMarker.M_SOS:
		         GetSos(dcInfo);
		              return true;

		      case JpegMarker.M_EOI:
		              return false;

		      default:
		              break;
		   }

		   return false;
	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * GetSos --
	 *
	 *      Process a SOS marker
	 *
	 * Results:
	 *      None.
	 *
	 * Side effects:
	 *      Bitstream is parsed.
	 *      Exits on error.
	 *
	 *--------------------------------------------------------------
	 */

	private int GetSos(DecompressInfo dcInfo) {
		int length;
		int i, n, c, cc;
		JpegComponentInfo jci = null;

		length = Get2bytes();

		if (length < 4) return -1;

		/*
		 * Get the number of image components.
		 */
		n = huffd.GetJpegChar();
		length -= 3;

		if (length != (n * 2 + 3) || n < 1 || n > 4) {
			System.err.println("Bogus SOS length");
			return -2;
		}

		for (i = 0; i < n; i++) {
			cc = huffd.GetJpegChar();
			c  = huffd.GetJpegChar();

			if (cc == dcInfo.getCompInfo().getComponentId()) {
				break;
			}

			jci = dcInfo.getCompInfo();
			jci.setDcTblNo((c >> 4) & 15);
		}

		/*
		 * Get the PSV, skip Se, and get the point transform parameter.
		 */
		dcInfo.setSs(huffd.GetJpegChar());
		c = huffd.GetJpegChar();
		c = huffd.GetJpegChar();
		dcInfo.setPt(c & 0x0F);

		return 0;		
	}

	/*
	 *--------------------------------------------------------------
	 *
	 * ReadFileHeader --
	 *
	 *      Initialize and read the file header (everything through
	 *      the SOF marker).
	 *
	 * Results:
	 *      -1 No SOI (FFD8) found.
	 *
	 * Side effects:
	 *      Exit on error.
	 *
	 *--------------------------------------------------------------
	 */
	int ReadFileHeader(DecompressInfo dcPtr)
	{
	   int c = 0;
	   int c2 = 0;

	   /*
	    * Demand an SOI marker (FFD8) at the start of the file --- otherwise
	    * it's probably not a JPEG file at all.
	    */
	   //c = huffd.GetJpegChar();
	   try {
		   c = inFile.read();
		   c2 = inFile.read();
	   } catch (IOException ioe) {
		   ioe.printStackTrace();
		   return -1;
	   }
	   //c2 = huffd.GetJpegChar();
	   if ((c != 0xFF) || (c2 != JpegMarker.M_SOI)) {
	           System.err.println("Not a JPEG file");
	           return -1;
	   } else {
		   String tmpStr = Integer.toHexString(c) + Integer.toHexString(c2);
		   System.err.println("Found SOI: " + tmpStr);
	   }

	   GetSoi(dcPtr);               /* OK, process SOI */

	   /*
	    * Process markers until SOF
	    */
	   c = ProcessTables(dcPtr);
	   if (c == JpegMarker.M_ERROR) return -2;

	   switch (c) {
	      case JpegMarker.M_SOF0:
	      case JpegMarker.M_SOF1:
	      case JpegMarker.M_SOF3:
	              GetSof(dcPtr);
	              break;

	      default:
	    	  	System.err.println("Unsupported SOF marker type 0x%02x: " + c);
	         	return -3;
	         	//break;
	   }

	   return 0;
	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * ProcessTables --
	 *
	 *      Scan and process JPEG markers that can appear in any order
	 *      Return when an SOI, EOI, SOFn, or SOS is found
	 *
	 * Results:
	 *      The marker found.
	 *
	 * Side effects:
	 *      Bitstream is parsed.
	 *
	 *--------------------------------------------------------------
	 */

	private int ProcessTables(DecompressInfo dcPtr) {

		int c;

		while (true) {
			c = NextMarker();
			if (c == JpegMarker.M_ERROR) return JpegMarker.M_ERROR;

			switch (c) {
			case JpegMarker.M_SOF0:
			case JpegMarker.M_SOF1:
			case JpegMarker.M_SOF2:
			case JpegMarker.M_SOF3:
			case JpegMarker.M_SOF5:
			case JpegMarker.M_SOF6:
			case JpegMarker.M_SOF7:
			case JpegMarker.M_JPG:
			case JpegMarker.M_SOF9:
			case JpegMarker.M_SOF10:
			case JpegMarker.M_SOF11:
			case JpegMarker.M_SOF13:
			case JpegMarker.M_SOF14:
			case JpegMarker.M_SOF15:
			case JpegMarker.M_SOI:
			case JpegMarker.M_EOI:
			case JpegMarker.M_SOS:
				return (c);

			case JpegMarker.M_DHT:
				GetDht (dcPtr);
				break;

			case JpegMarker.M_DQT:
				System.err.println("Not a lossless JPEG file.");
				return JpegMarker.M_ERROR;

			case JpegMarker.M_DRI:
				GetDri (dcPtr);
				break;

			case JpegMarker.M_APP0:
				GetApp0();
				break;

			case JpegMarker.M_RST0:              /* these are all parameterless */
			case JpegMarker.M_RST1:
			case JpegMarker.M_RST2:
			case JpegMarker.M_RST3:
			case JpegMarker.M_RST4:
			case JpegMarker.M_RST5:
			case JpegMarker.M_RST6:
			case JpegMarker.M_RST7:
			case JpegMarker.M_TEM:
				System.err.println("Warning: unexpected marker 0x%02x: " + c);
				break;

			default:          /* must be DNL, DHP, EXP, APPn, JPGn, COM, * or RESn */
				SkipVariable();
			break;
			}
		}

	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * NextMarker --
	 *
	 *      Find the next JPEG marker Note that the output might not
	 *      be a valid marker code but it will never be 0 or FF
	 *
	 * Results:
	 *      The marker found.
	 *
	 * Side effects:
	 *      Bitstream is parsed.
	 *
	 *--------------------------------------------------------------
	 */

	private int NextMarker() {
		int c;
		int bytecnt = 0;

		do {
			/*
			 * skip any non-FF bytes
			 */
			do {
				c = huffd.GetJpegChar();
				bytecnt++;
				if (bytecnt > 50000) return JpegMarker.M_ERROR;
			} while (c != 0xFF);

			/*
			 * skip any duplicate FFs without incrementing nbytes, since
			 * extra FFs are legal
			 */
			do {
				c = huffd.GetJpegChar();
				bytecnt++;
				if (bytecnt > 50000) return JpegMarker.M_ERROR;
			} while (c == 0xFF);

		} while (c == 0);             /* repeat if it was a stuffed FF/00 */

		return c;

	}

/*	private int GetJpegChar() {
		int c = 0;
		try {
			c = inFile.read();
		} catch (IOException ioe) {
			// TJJ XXX how to handle this?
			System.err.println("ERROR: Unexpected EOF!");
		}
		return c;
	}*/

	private void GetSoi(DecompressInfo dcPtr) {
		dcPtr.setRestartInterval(0);
	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * Get2bytes --
	 *
	 *      Get a 2-byte unsigned integer (e.g., a marker parameter length
	 *      field)
	 *
	 * Results:
	 *      Next two byte of input as an integer.
	 *
	 * Side effects:
	 *      Bitstream is parsed.
	 *
	 *--------------------------------------------------------------
	 */
	int Get2bytes()
	{
	    int a;

	    a = huffd.GetJpegChar();
	    return (a << 8) + huffd.GetJpegChar();
	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * GetDri --
	 *
	 *      Process a DRI marker
	 *
	 * Results:
	 *      None
	 *
	 * Side effects:
	 *      Exits on error.
	 *      Bitstream is parsed.
	 *
	 *--------------------------------------------------------------
	 */
	int GetDri(DecompressInfo dcPtr)
	{
	   if (Get2bytes() != 4) {
	      System.err.println("Bogus length in DRI");
	           //exit (1);
	      return -1;
	   }

	   dcPtr.setRestartInterval(Get2bytes());
	   //dcPtr->restartInterval = Get2bytes();

	   return 0;
	}

	/*
	 *--------------------------------------------------------------
	 *
	 * GetApp0 --
	 *
	 *      Process an APP0 marker.
	 *
	 * Results:
	 *      None
	 *
	 * Side effects:
	 *      Bitstream is parsed
	 *
	 *--------------------------------------------------------------
	 */
	void GetApp0()
	{
	   int length;

	   length = Get2bytes() - 2;
	   while (length-- > 0) {     /* skip any remaining data */
	      huffd.GetJpegChar();
	   }
	}

	/*
	 *--------------------------------------------------------------
	 *
	 * SkipVariable --
	 *
	 *      Skip over an unknown or uninteresting variable-length marker
	 *
	 * Results:
	 *      None.
	 *
	 * Side effects:
	 *      Bitstream is parsed over marker.
	 *
	 *
	 *--------------------------------------------------------------
	 */
	void SkipVariable()
	{
	   int length;

	   /* Set length to marker length minus the two bytes of the header length. */
	   length = Get2bytes() - 2;

	   /* Simply skip over the current marker. */
	   while (length-- >= 0) {
	      huffd.GetJpegChar();
	   }
	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * GetSof --
	 *
	 *      Process a SOFn marker
	 *
	 * Results:
	 *      None.
	 *
	 * Side effects:
	 *      Bitstream is parsed
	 *      Exits on error
	 *      DecompressInfo object is filled in
	 *
	 *--------------------------------------------------------------
	 */
	
	int GetSof(DecompressInfo di)
	{
		int length;
		int c;
		JpegComponentInfo jci = null;

		length = Get2bytes();

		if (length < 6) return -1;

		di.setDataPrecision(huffd.GetJpegChar());
		di.setImageHeight(Get2bytes());
		di.setImageWidth(Get2bytes());
		// just skip the fixed-val num components field
		huffd.GetJpegChar();

		/*
		 * We don't support files in which the image height is initially
		 * specified as 0 and is later redefined by DNL.  As long as we
		 * have to check that, might as well have a general sanity check.
		 */
		if ((di.getImageHeight() <= 0 ) || (di.getImageWidth() <= 0)) {
			System.err.println("Empty JPEG image (DNL not supported)");
			return -2;
		}

		if ((di.getDataPrecision() < MinPrecisionBits) ||
				(di.getDataPrecision() > MaxPrecisionBits)) {

			System.err.println("Unsupported JPEG data precision");
			return -3;
		}
		
		if (length != (3 + 8)) {
			System.err.println("Bogus SOF length");
			return -4;
		}

		jci = new JpegComponentInfo();
		jci.setComponentIndex(0);
		jci.setComponentId(huffd.GetJpegChar());
		c = huffd.GetJpegChar();
		jci.setHSampFactor((c >> 4) & 15);
		jci.setVSampFactor((c) & 15);
		huffd.GetJpegChar();   /* skip Tq */

		di.setCompInfo(jci);

		return 0;
	}
	
	/*
	 *--------------------------------------------------------------
	 *
	 * GetDht --
	 *
	 *      Process a DHT marker
	 *
	 * Results:
	 *      Negative error codes on failure.
	 *
	 * Side effects:
	 *      A huffman table is read.
	 *      Exits on error.
	 *
	 *--------------------------------------------------------------
	 */
	
	int GetDht(DecompressInfo di)
	{
		int length;
		byte [] bits = new byte[17];
		byte [] huffval = new byte[256];
		int i, index, count;
		HuffmanTable ht;

		length = Get2bytes() - 2;

		if (length < 1) return -1;

		while (length != 0) {
			index = huffd.GetJpegChar();

			bits[0] = 0;
			count = 0;
			for (i = 1; i <= 16; i++) {
				bits[i] = (byte) huffd.GetJpegChar();
				count += (int) (bits[i] & 0xFF);
			}

			if (count > 256) {
				System.err.println("Bogus DHT counts");
				return -2;
			}

			for (i = 0; i < count; i++) {
				huffval[i] = (byte) huffd.GetJpegChar();
			}

			length -= 1 + 16 + count;

			if ((index & 0x10) != 0) {        /* AC table definition */
				System.err.println("Huffman table for lossless JPEG is not defined.");
				return -3;
			}
			else {            /* DC table definition */
				ht = di.getDcHuffTbl();
			}

			if (index < 0 || index >= 4) {
				System.err.println("Bogus DHT index: " + index);
				return -4;
			}

			ht.setBits(bits);
			ht.setHuffval(huffval);
			di.setDcHuffTbl(ht);
		}

		return 0;
	}
      
}
