package nasa.stps.core.xstps.hrit;

import nasa.stps.core.Builder;
import nasa.stps.core.StpsNode;

import nasa.stps.core.ccsds.Packet;
import nasa.stps.core.status.StatusItem;

import nasa.stps.server.SDIServer;

import java.io.*;

import java.text.SimpleDateFormat;
import java.util.Date;

/**
 * This class manages the output files. It deals with identifying image segments,
 * invoking a decoder for the lossless JPEG data, and writing out data files.
 * @author Tommy Jasmin, May 2006
 */

class OutputFiles
{
    private int PRIMARY_HEADER_LENGTH = 16;
    private static final byte FILE_TYPE_IMAGE_DATA = 0;
    private static final byte FILE_TYPE_TEXT_DATA = 1;
    private static final int NUM_SEGMENT_FILES_FULL_DISK = 50;
    private static final int NUM_SEGMENT_FILES_HALF_DISK = 25;
    public static final String IN_PROGRESS_PREFIX = "INCOMING_";
    private String curFileName = null;
    private String prvFileName = null;
    // this is used to help determine end of image
    private int curSegmentCount = 0;
    private String path;
    // regular expression for files to purge. default is image files,
    // will most likely be overridden by a system property
    private String purgePattern = "IMG_.*";
    private OutputStream hritOut;
    private PipedOutputStream jpegOut;
    private boolean imageSegmentInProgress = false;
    // Note the difference here, below flag is for ENTIRE image, not just a segment
    private boolean imageInProgress = false;
    private boolean textMessageInProgress = false;
    private int[] xappid = new int[3];
    private int[] xspid = new int[3];
    static int lastPacketDataLength = 0;
    // default expire time: 3 days * 24 hr/day * 60 min/hr * 60 sec/min * 1000 ms/sec
    long expireTimeMillis = 3 * 24 * 60 * 60 * 1000;
    // date/time formatter for image start/end events
    SimpleDateFormat sdf = new SimpleDateFormat("yyyy.DDD.HHmmss");
    // date/time formatter for MANAM
    SimpleDateFormat sdfMANAM = new SimpleDateFormat("yyyyMMDDHHmm");
    // path to notification binary should be a system prop, set below...
    String notify = "";
    // whether or not to write satellite name as ancillary header in output files
    boolean tagSatName = false;
    // spacecraft name, prepended to any output file
    private String spacecraftName = null;
    // separate JPEG decoder threads will run serially, so only need one
    Thread t = null;
    // Builder reference for logging stats at end of each image
    private Builder builder;
    
    // image filename prefixes
    // full disk
    private static final String FULL_DISK = "DK01";
    // northern hemisphere
    private static final String NORTHERN_HEMI = "DK02";
    // southern hemisphere
    private static final String SOUTHERN_HEMI = "DK03";

    /**
     * Create an OutputFiles object.
     * @param path The directory where it will put the data files
     */
    
    OutputFiles(String path) throws IOException
    {
        this.path = path;
        // see if a non-default data expiration was set via System Property
        String expireStr = System.getProperty("expire_days");
        if (expireStr != null) {
        	// try to convert to an integer
        	try {
        		int expireDays = Integer.parseInt(expireStr);
        		// if successful conversion, use the supplied value
        		expireTimeMillis = expireDays * 24 * 60 * 60 * 1000;
        	} catch (NumberFormatException nfe) {
        		// do nothing, we'll just leave the default setting
        	}
        }
        // see if a notification binary, with absolute path, was set
        String tmpNotify = System.getProperty("notify_with_abs_path");
        if (tmpNotify != null) {
        	notify = tmpNotify;
        }
        String tmpTagSatName = System.getProperty("tag_sat_name", "false");
        if (tmpTagSatName != null) {
        	if (tmpTagSatName.equalsIgnoreCase("true")) {
        		tagSatName = true;
        	}
        }
        // see if there is a regular expression specified for file purges
        String tmpPurgePattern = System.getProperty("purge_regex");
        if (tmpPurgePattern != null) {
        	purgePattern = tmpPurgePattern;
        }
    }

    /**
     * See if this packet has an annotation header (starts a new file).
     * @param packet The current packet being processed
     */
    
    private String annotationHeaderPresent(Packet packet)
    {
    	// is total header length more than primary header length?  
    	// (i.e., multiple headers exist in this packet)
    	if (packet.secondaryHeaderFlagSet()) {
    		System.out.println("\nLooks like start of a new file... Header type: " + packet.getFirstDataByte());
    		System.out.println("File type (0: image, 2: text) : " + packet.getDataByteAtOffset(3));
    		if (packet.totalHeaderLength() != PRIMARY_HEADER_LENGTH) {   			
    			return packet.getAnnotationText();
    		} else {
    			return null;
    		}
    	} else {
    		return null;
    	}
    }  

    /**
     * Get the current file name
     */
    
    final String getFileName()
    {
    	return curFileName;
    }

    /**
     * Get the directory where this class is putting data files.
     */
    
    final String getPath()
    {
        return path;
    }
    
    /**
     * Log and clear statistics accumulated since last image. 
     * This is where we currently dump periodic stats for all nodes,
     * such as bad frames, detected bit errors, etc. 
     */
    
    private void logStats() {
    	if (builder != null) {
			// get a reference to all existing nodes
			java.util.Map nodes = builder.getConfiguration().getStpsNodes();
			java.util.Iterator i = nodes.values().iterator();
			System.out.println("\nProcessing statistics: (counts reset each time logged)");
			while (i.hasNext())
			{
				StpsNode node = (StpsNode)i.next();
				java.util.Collection status = node.getStatusItems();
				if (status != null)
				{
					System.out.println(node);
					java.util.Iterator ii = status.iterator();
					while (ii.hasNext())
					{
						StatusItem si = (StatusItem) ii.next();
						System.out.println("    " + si);
						si.clear();
					}
				}
			}
			System.out.println("Status counts cleared.");
    	}
    }

    /**
     * Set the application id and spacecraft id for a specific application
     * index. The index is an array index (0,1,2) and not an application id.
     * @param apindex
     * @param appid
     * @param spid
     */
    
    final void setAppidSpid(int apindex, int appid, int spid)
    {
        xappid[apindex] = apindex;
        xspid[apindex] = spid;
    }
    
    /**
     * Set the Builder
     * @param b
     */
    
    final void setBuilder(Builder b)
    {
    	builder = b;
    }
    
    /**
     * Set the spacecraft name, comes from element attribute in XML config file
     * @param s
     */
    
    final void setSpacecraftName(String s)
    {
    	spacecraftName = s;
    }
    
    /**
     * Open a segment file on disk
     * @param fileType
     * @throws FileNotFoundException
     */
    
    private void openFile(byte fileType) throws FileNotFoundException
    {
        // close any currently open streams.  this will allow any potentially stuck
    	// threads (in the very unlikely event of a missed final packet) to complete
    	try {
    		if (hritOut != null) {
    			hritOut.flush();
    			hritOut.close();
    		}
    		if (jpegOut != null) {
    			jpegOut.flush();
    			jpegOut.close();
    		}
    	} catch (IOException ioe) {
    		System.err.println("Cleaning up old streams before opening new data file...");
    		ioe.printStackTrace();
    	}
        
        // actual file on disk gets a INCOMING_ prepended, removed when image is complete
        File hritFile = new File(path, IN_PROGRESS_PREFIX + curFileName);
        FileOutputStream fos = new FileOutputStream(hritFile);
        hritOut = new BufferedOutputStream(fos, 8192);
        // only create data pipe (of compressed JPEG data) for image files
        if (fileType == FILE_TYPE_IMAGE_DATA) {
        	jpegOut = new PipedOutputStream();
        	imageSegmentInProgress = true;
			curSegmentCount++;
			System.out.println("Current segment count for this image: " + curSegmentCount);
        } else {
        	textMessageInProgress = true;
        }
    }
    
    /**
     * Write a message to the system console.
     * @param message
     */
    
    private void writeToConsole(String message) {
        try {
           PrintWriter pw = new PrintWriter(new FileOutputStream("/dev/console"), true);
           pw.println(message);
           pw.close();
        }
        catch (FileNotFoundException fnfe) {
           System.out.println("Cannot write to system console");
        }
    }

    /**
     * Write a packet to a data file.
     * @param packet
     * @param apindex
     * @throws IOException
     */
    
    void write(Packet packet, int apindex)
            throws IOException
    {    	
    	// action depends on sequence flag, will be either start of 
    	// image, end of image, self-contained, or continuation of current image
    	int sequenceFlag = packet.getSequenceFlags();
    	switch (sequenceFlag) {
    	
    		// packet contains one complete file
    		case 3:
    			// write these out as text files...
    			// create a current date/time string
    			Date rightNow = new Date();
    			String dateTimeString = sdfMANAM.format(rightNow);
    			
    			// prepend MANAM to form filename, and hope this *is* MANAM!
    			String textFileName = "MANAM_" + dateTimeString;
    			File textFile = new File(path, textFileName);
    			FileOutputStream fos = new FileOutputStream(textFile);
    			System.out.println("\nProcessing file as text message: " + textFileName + ", apindex: " + apindex);
    			System.out.println("\nTotal header length this packet: " + packet.totalHeaderLength());
    			// write out complete file
    			fos.write(packet.getData(), 
    					Packet.PRIMARY_HEADER_LENGTH + Packet.TRANSPORT_HEADER_LENGTH + packet.totalHeaderLength(), 
    					packet.getData().length - Packet.PRIMARY_HEADER_LENGTH - Packet.TRANSPORT_HEADER_LENGTH - packet.totalHeaderLength());    			
    			// type 3 is self-contained, so close it out now!
            	fos.flush();
            	fos.close();
            	
            	// save it a second time as the static filename "MANAM"
    			File staticFile = new File(path, "MANAM");
    			// probably don't need to stream this, but to be consistent with image output mechanism...
    			FileOutputStream fos2 = new FileOutputStream(staticFile);
    			// write out complete file
    			fos2.write(packet.getData(), 
    					Packet.PRIMARY_HEADER_LENGTH + Packet.TRANSPORT_HEADER_LENGTH + packet.totalHeaderLength(), 
    					packet.getData().length - Packet.PRIMARY_HEADER_LENGTH - Packet.TRANSPORT_HEADER_LENGTH - packet.totalHeaderLength());    			
    			// type 3 is self-contained, so close it out now!
            	fos2.flush();
            	fos2.close();            	
            	
				// we also do event notifications for each text message
				sendNotification(textFileName, "manam");
    			break;
    		
	    	// start of new file
	    	case 1:
	    		// set previous file name, in case we need to reference it for missed
	    		// EOI segments (in which case we force an EOI event for prev image)
	    		if (curFileName != null) {
	    			prvFileName = curFileName;
	    		}
	    		// pull out file name
	    		curFileName = annotationHeaderPresent(packet);
	    		if (curFileName == null) {
	    			System.err.println("Error: unable to determine name for new file");
	    		} else {
	    			// if it's an image or overlay file
	    			if ((curFileName.startsWith("IMG_")) || (curFileName.startsWith("OVL_"))) {
	    				System.out.println("\nFound start of a new file! Filename: " + curFileName);
	    				
	    				// first, make previous JPEG decoding threads are complete
	    				// give previously created threads up to 30 seconds to finish up
	    				// in practice it should need at most 15 seconds so this is padded big
	    				// NOTE: t will only be null when the program first starts up
	    				if (t != null) {
	    					try {
	    						System.err.println("Waiting 45 seconds for previous JPEG thread to finish...");
	    						t.join(45000);
	    						// see if timeout occurred and thread is still alive
	    						if (t.isAlive()) {
	    							System.err.println("HRITJpegDecoder thread join timed out! Old thread still alive, will try to kill and clean up...");
	    							// do something - try to clean up and kill old thread
	    							// first, close the pipe in case it was hung on I/O
	    							if (jpegOut != null) {
	    								jpegOut.flush();
	    								jpegOut.close();
	    							}
	    			    			// then, pause a bit in case we unstuck it, give it a chance to release resources
	    			    			Thread.sleep(5000);
	    			    			// finally, null it out
	    			    			t = null;
	    						}
	    					} catch (InterruptedException ie) {
	    						ie.printStackTrace();
	    						t.interrupt();
	    					}
	    				}
	    				
	    				openFile(FILE_TYPE_IMAGE_DATA);	 
	    				// store original length of all headers, needed to write out that chunk of data, and 1st jpg chunk
	    				int oldLen = packet.totalHeaderLength();
	    				// write out header info, if config flag indicates, add header with text satellite name
	    				if (tagSatName) {
	    					// new length is old plus new header (type field 1b, length field 2b, body Xb)
	    					int newLen = oldLen + 1 + 2 + spacecraftName.length();
	    					// but in the data written, reflect the additional header we will tack on
	    					packet.setTotalHeaderLength(newLen);
	    					// write original headers, which reference fact that there will be an additional one
	    					hritOut.write(packet.getData(), Packet.PRIMARY_HEADER_LENGTH + Packet.TRANSPORT_HEADER_LENGTH, oldLen);
	    					// finally, create and write the new header, which is a Header Type #6
	    					// (Ancillary Text) which stores the satellite name
	    					// size is header type (1 byte) plus header length (2 bytes) plus text
	    					byte [] extraHeader = new byte[1 + 2 + spacecraftName.length()];
	    					extraHeader[0] = (byte) 6;
	    					extraHeader[2] = (byte) (spacecraftName.length() + 3);
	    					byte [] asciiSpacecraftName = spacecraftName.getBytes("US-ASCII");
							System.err.println("Tagging spacecraft name: " + spacecraftName);
	    					if (asciiSpacecraftName != null) {
	    						for (int idx = 0; idx < asciiSpacecraftName.length; idx++) {
	    							extraHeader[3 + idx] = asciiSpacecraftName[idx];
	    						}
	    					}
	    					hritOut.write(extraHeader, 0, extraHeader.length);
	    				} else {
	    					hritOut.write(packet.getData(), Packet.PRIMARY_HEADER_LENGTH + Packet.TRANSPORT_HEADER_LENGTH, packet.totalHeaderLength());
	    				}
	    				
	    				// create the thread that will decode the JPEG data on the fly
	    				// pass in length of all headers as offset to start of data
	    				t = new HRITJpegDecoder(jpegOut, IN_PROGRESS_PREFIX + curFileName, packet.totalHeaderLength(), path);
	    				t.start();

	    				// write out jpg data
	    				jpegOut.write(packet.getData(), 
	    						Packet.PRIMARY_HEADER_LENGTH + Packet.TRANSPORT_HEADER_LENGTH + oldLen, 
	    						packet.getDataFieldLength() - Packet.TRANSPORT_HEADER_LENGTH - oldLen - 2);
	    			// otherwise we'll consider it a text file
	    			} else {
	    				openFile(FILE_TYPE_TEXT_DATA);
	    				System.out.println("\nProcessing file as text message: " + curFileName);
	    			}
	    		}
	    		// do external event notifications
    			// What should the logic be for SOI?  For now, checking segment count for the image
	    		if (imageSegmentInProgress) {
	    			if (curSegmentCount == 1) {
	    				// before any image start events, make sure there are no prior images
	    				// that were not "closed out" properly with an image end event (this 
	    				// could happen if final segment was noisy and lost for ex)
	    				if (imageInProgress) {
	    					if (prvFileName != null) {
	    						sendNotification(prvFileName, "imageend");
	    					}
	    				}
	    				sendNotification(curFileName, "imagestart");
	    	    		// this is the place to set the overall image in progress flag
	    	    		imageInProgress = true;
	    			}	
	    		}
	    		break;
	    	
	    	// end of file
	    	case 2:    		
	    		// write out last bit of data - don't include the two CRC bytes at the end...
	    		if (imageSegmentInProgress) {
	    			try {
	    				jpegOut.write(packet.getData(), Packet.PRIMARY_HEADER_LENGTH, packet.getDataFieldLength() - 2);
	    			} catch (IOException ioe) {
	    				System.err.println("Write to JPEG thread failed, noisy data? decoder finished early?");
	    				// don't try and send any more data out the pipe, since it isn't there!
	    				imageSegmentInProgress = false;
	    			}
	    			// close current output streams, won't need them now that we're writing final data out
	    			close();
	    		}
	    		
	    		// text message packets are easier, just write the data out
	    		if (textMessageInProgress) {
	    			hritOut.write(packet.getData(), Packet.PRIMARY_HEADER_LENGTH, packet.getDataFieldLength() - 2);
	    			close();
	    		}
	    		
	    		// at each end-of-image, check for and purge any expired image files,
	    		// and send out an End-Of-Image event
	    		if (curFileName != null) {
	    			// if so, if it's a full-disk, segment count should be 50, for half-disks, it should be 25
	    			if (((curFileName.indexOf(FULL_DISK) > 0) && (curSegmentCount == NUM_SEGMENT_FILES_FULL_DISK)) ||
	    			   (((curFileName.indexOf(NORTHERN_HEMI) > 0) || (curFileName.indexOf(SOUTHERN_HEMI) > 0)) && (curSegmentCount == NUM_SEGMENT_FILES_HALF_DISK))) {
	    				File imageDirectory = new File(path);
	    				File[] directoryContents = imageDirectory.listFiles();
	    				long currentDate = System.currentTimeMillis();
	    				if (directoryContents != null) {
	    					int count = directoryContents.length;
	    					// loop through contents of image directory
	    					for (int i = 0; i < count; i++) {
	    						File f = directoryContents[i];
	    						// make sure it's a file!  and one that is specified to be purged...
	    						if ((f.isFile()) && (f.getName().matches(purgePattern))) {
	    							if (f.lastModified() < (currentDate - expireTimeMillis)) {
	    								System.out.println("Purging file: " + f.getName());
	    								f.delete();
	    							}
	    						}
	    					}
	    				}
	    				// we also do event notifications at end-of-image
	    				sendNotification(curFileName, "imageend");
	    				imageInProgress = false;
	    				curSegmentCount = 0;
	    				// and log, then clear statistics
	    				logStats();
	    				// finally, rebuild the entire processing pipeline after 1/2 or full disks to free up memory
	    				if (curFileName.indexOf("DK0") > 0) SDIServer.running = false;
	    			}
	    		}
	    		break;
	    	
	    	// continuation of current file
	    	case 0:
	    		// don't include the two CRC bytes at the end...
	    		if (imageSegmentInProgress) {
	    			try {
	    				jpegOut.write(packet.getData(), Packet.PRIMARY_HEADER_LENGTH, packet.getDataFieldLength() - 2);
	    			} catch (IOException ioe) {
	    				System.err.println("Write to JPEG thread failed, noisy data? decoder finished early?");
	    				// don't try and send any more data out the pipe, since it isn't there!
	    				imageSegmentInProgress = false;
	    			}
	    		}
	    		if (textMessageInProgress) {
	    			hritOut.write(packet.getData(), Packet.PRIMARY_HEADER_LENGTH, packet.getDataFieldLength() - 2);
	    		}	    		
	    		break;
	    	
	    	// not supposed to be any other possibilities
	    	default:
	    		System.err.println("Error: Unknown file type: " + sequenceFlag);
	    		break;
    		
    	}
        
    }

    /**
     * Close the current data file.
     * @throws IOException
     */

    void close() throws IOException
    {
        // close any currently open streams
        if (hritOut != null) {
        	hritOut.flush();
        	hritOut.close();
        }
        
        if (jpegOut != null) {
        	jpegOut.flush();
        	jpegOut.close();
        }
        
        // reset in-progress flags, why check which type?, easier to just reset 'em both
        imageSegmentInProgress = false;
        textMessageInProgress = false;
    }    
    
    /**
     * Do the external event notifications for start, end of image.
     * @param filename
     * @param event
     */
    
    void sendNotification(String filename, String event)
    {
    	// MANAM (text - "manual amendment" events)
    	if (event.equals("manam")) {
    		// create a current date/time string
    		Date rightNow = new Date();
    		String dateTimeString = sdf.format(rightNow);
    		// send the event
    		try {
        		String [] cmdData = new String[3];
        		cmdData[0] = notify;
        		cmdData[1] = event;
        		cmdData[2] = filename;
     			Runtime.getRuntime().exec(cmdData);
    			// also log the event
    			System.out.println(dateTimeString + ":" + filename + " manam");
    			writeToConsole(dateTimeString + ":" + filename + " manam");
    		} catch (IOException ioe) {
    			System.err.println("manam notification failed, notify binary missing?");
    		}
    	}    	
    	// start of image events
    	if (event.equals("imagestart")) {
    		// create a current date/time string
    		Date rightNow = new Date();
    		String dateTimeString = sdf.format(rightNow);
    		// send the event
    		try {
        		String [] cmdData = new String[3];
        		cmdData[0] = notify;
        		cmdData[1] = event;
        		cmdData[2] = dateTimeString + ":" + filename + " start";
     			Runtime.getRuntime().exec(cmdData);
    			// also log the event
    			System.out.println(dateTimeString + ":" + filename + " start");
    			writeToConsole(dateTimeString + ":" + filename + " start");
    		} catch (IOException ioe) {
    			System.err.println("imagestart notification failed, notify binary missing?");
    		}
    	}
    	// end of image events, very similar
    	if (event.equals("imageend")) {
    		// create a current date/time string
    		Date rightNow = new Date();
    		String dateTimeString = sdf.format(rightNow);    			
    		// send the event
    		try {
        		String [] cmdData = new String[3];
        		cmdData[0] = notify;
        		cmdData[1] = event;
        		cmdData[2] = dateTimeString + ":" + filename + " end";
     			Runtime.getRuntime().exec(cmdData);
    			// also log the event
    			System.out.println(dateTimeString + ":" + filename + " end");
    			writeToConsole(dateTimeString + ":" + filename + " end");
    		} catch (IOException ioe) {
    			System.err.println("imageend notification failed, notify binary missing?");
    		}
    	}
    }
    
}
