package nasa.stps.server;

import nasa.stps.core.fs.FrameSynchronizer;
import nasa.stps.core.Builder;
import nasa.stps.core.StpsException;
import nasa.stps.core.StpsNode;

import java.io.FileReader;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;

/**
 * This is an abstract STPS server. It supports only one pipeline thread.
 * @author Stan Hilinski, September, 2001
 */

public abstract class AbstractServer implements StpsServices, Runnable
{
    protected FrameSynchronizer fs = null;
    public static boolean running = true;
    protected Thread runner;
    protected StatusBlock[] statusBlocks = null;
    protected ServerState state;
    protected Builder builder;
    private Logger logger;
    private String xmlDirectory;

    /**
     * Create an STPS server. This class contains core server functions.
     */
    protected AbstractServer()
    {
        super();

        System.out.println(nasa.stps.Version.getVersion());

        //Identify the configuration file directory.
        xmlDirectory = System.getProperty("setup", "config");

        //The logger records all events.
        logger = new Logger();

        //The server state tracks load/unloaded and stop/go.
        state = new ServerState();

        //The builder creates pipelines.
        try
        {
            builder = new Builder();
        }
        catch (nasa.stps.core.StpsException stpse)
        {
            stpse.printStackTrace();
            System.exit(0);
        }
    }

    /**
     * Log a message.
     */
    public final void log(String message)
    {
        logger.print(message);
    }

    /**
     * Close the log.
     */
    public final void closeLog()
    {
        logger.close();
    }

    /**
     * Get the name of the currently loaded configuration.
     * @return null if none is loaded.
     */
    public synchronized String getConfigurationName()
    {
        String name = null;
        if (state.isLoaded())
        {
            name = builder.getConfiguration().getName();
        }
        return name;
    }

    /**
     * Is the server loaded and enabled?
     */
    public boolean isEnabled()
    {
        return state.isEnabled();
    }

    /**
     * Enable or disable data processing.
     */
    public synchronized void setEnabled(boolean enable)
    {
        if (enable)
        {
            log("Go");
            state.go();
        }
        else
        {
            log("Stop");
            state.stop();
        }
    }

    /**
     * Unload a configuration.
     * @throws StpsException 
     */
    public synchronized void unload() throws StpsException
    {
        if (state.isLoaded())
        {
            log("Unload " + getConfigurationName());
            try
            {
                state.unload();
                fs.shutdown();
            }
            catch (nasa.stps.core.StpsException stpse)
            {
                log("Unload error " + stpse.getMessage());
                throw stpse;
            }
            finally
            {
                fs = null;
            }
        }
        else
        {
            log("Unload commanded, but the server is not loaded.");
        }
    }

    /**
     * Load a configuration.
     * @throws IOException 
     * @throws StpsException 
     */
    public synchronized void load(String configuration) throws StpsException, IOException
    {
        log("Load configuration as a string");
        build(new java.io.StringReader(configuration));
    }

    /**
     * Load a configuration from a file that is local to the server.
     * @throws IOException 
     * @throws StpsException 
     */
    public synchronized void loadFile(String filename) throws StpsException, IOException
    {
        log("Load " + filename);
        File setupFile = new File(xmlDirectory,filename);
        BufferedReader r = null;
        try
        {
            r = new BufferedReader(new FileReader(setupFile));
        }
        catch (java.io.FileNotFoundException fnfe)
        {
            log("Load error " + fnfe.getMessage());
            throw fnfe;
        }
        build(r);
    }

    /**
     * Build a pipeline from a configuration.
     * @throws StpsException 
     * @throws IOException 
     */
    private void build(java.io.Reader r) throws StpsException, IOException
    {
        //The server must be unloaded first.
        if (state.isLoaded())
        {
            unload();
        }

        try
        {
            //The builder constructs a pipeline from the configuration.
            org.xml.sax.InputSource setup = new org.xml.sax.InputSource(r);
            String root = System.getProperty("root","file://localhost/");
            setup.setSystemId(root);
            fs = builder.create(setup);
            r.close();

            //I count the number of status blocks.
            java.util.Map nodes = builder.getConfiguration().getStpsNodes();
            java.util.Iterator i = nodes.values().iterator();
            int count = 1;  //1 for the server status
            while (i.hasNext())
            {
                StpsNode node = (StpsNode)i.next();
                if (node.getStatusItems() != null) ++count;
            }

            //I create a status block array. The server state is first.
            statusBlocks = new StatusBlock[count];
            statusBlocks[0] = state.getStatusBlock();

            //I put all status blocks into the array.
            count = 1;
            i = nodes.values().iterator();
            while (i.hasNext())
            {
                StpsNode node = (StpsNode)i.next();
                if (node.getStatusItems() != null)
                {
                    statusBlocks[count++] = new StatusBlock(node);
                }
            }

            //Update the state.
            state.load(builder.getConfiguration().getName());
        }
        catch (nasa.stps.core.StpsException stpse)
        {
            log("Load error " + stpse.getMessage());
            stpse.printStackTrace();
            throw stpse;
        }
        catch (java.io.IOException ioe)
        {
            log("Load error " + ioe.getMessage());
            throw ioe;
        }
    }

    /**
     * Stop the STPS server.
     */
    public void stopServer()
    {
        running = false;
        try { runner.join(3000L); }
        catch(InterruptedException ie1) {  }
        log("Someone stopped the server.");
        System.exit(0);
    }

    /**
     * Set all clearable status counters to zero.
     */
    public synchronized void zeroStatus()
    {
        if (statusBlocks != null)
        {
            for (int n = 0; n < statusBlocks.length; n++)
            {
                statusBlocks[n].clear();
            }
            log("Zero status counts.");
        }
    }

    /**
     * Get the current status from the last-used pipeline. Within a session,
     * the number and identity of the status blocks and the number and
     * identity of the contained status items will not change. Only the
     * status item values may change. However, when a session changes and
     * a new configuration is loaded, then all of these elements will change.
     * @return If no configuration has been loaded into the server, then
     * this method returns null. Otherwise it returns the status of the
     * last created pipeline, even post-session.
     */
    public synchronized StatusBlock[] getStatus()
    {
        state.advanceClock();
        return statusBlocks;
    }


    /**
     * This thread method gets data from the input socket and sends it down the
     * pipeline if it exists. It discards the data if the pipeline is null.
     */
    public abstract void run();
}
