package nasa.stps.server;

import java.io.*;
import java.text.SimpleDateFormat;
import java.util.Date;

import nasa.stps.core.StpsException;

/**
 * This is an STPS server that receives telemetry through an SSEC SDI card.
 * <p>
 * You may pass one argument when invoking, the server name. The full server
 * name then becomes "StpsServices."+name. The default full server name is
 * "StpsServices.A".
 * <p>
 * The server understands the following system properties. You set them when
 * you run the server using the -D attribute as in "-Dport=5000".
 * <br>
 * -Dconfig=xmlConfigFileName
 * <br>A configuration file to be used until overridden by a loaded one. The
 * default name is "default.xml".
 * <br>
 * -Ddevice=/dev/devicename
 * <br>The device/card that the server reads for telemetry data. The
 * default is /dev/sib0.
 * <br>
 * -DbufferSizeKb=8
 * The amount of data I accumulate before processing. The default is 256 KB.
 * <br>
 * -Dsetup=configurationDirectory
 * <br>The directory where local configuration files are found. If provided,
 * all files must be within the directory tree.
 * @author Stan Hilinski, April, 2001, modified by Tommy Jasmin, June 2006
 */

public class SDIServer extends AbstractServer
{
    private int inputBufferSize = 4 * 1024;
    private static String defaultDevice = "/dev/sib0";
    private String deviceName = null;
    private String notify = "";
    private static final int CONSECUTIVE_READ_ERROR_LIMIT = 100;
    // date/time formatter for ingestor restart events
    SimpleDateFormat sdf = new SimpleDateFormat("yyyy.DDD.HHmmss");

    /**
     * If a data source connects to this server and there is no loaded
     * configuration, the server will automatically load one of the two
     * following configurations. One will be null. When a user loads a
     * new configuration, it also becomes the new automatic default.
     */
    private String defaultConfigurationFileName = null;

    /**
     * This is the main entry point for the STPS server.
     */
    public static void main(String args[])
    {
        // XXX TJJ if we no longer need server name, config file can probably change
    	//String serverName = (args.length == 0)? "SDI" : args[0];
        String defaultConfig = System.getProperty("config", "default.xml");
        // see if a notification binary, with absolute path, was set
        String notifyWAP = System.getProperty("notify_with_abs_path");
        String device = null;
        int bufferSize = 0;
        String propertyName = null;

        try
        {
            //Get the device driver location system property.
            propertyName = "device";
            device = System.getProperty(propertyName, defaultDevice);

            //Get the input buffer size.
            propertyName = "bufferSizeKb";
            String x = System.getProperty(propertyName, "256");
            bufferSize = 1024 * Integer.valueOf(x).intValue();
        }
        catch (NumberFormatException nfe)
        {
            System.err.println(propertyName + " property is not an integer.");
            System.exit(-3);
        }

        try
        {
            SDIServer server = new SDIServer(defaultConfig, device, bufferSize, notifyWAP);
            server.log("Ready to process data.");
        }
        catch (Exception ex)
        {
            System.err.println("SERVER ERROR. " + ex);
            ex.printStackTrace();
            System.exit(-2);
        }
    }

    /**
     * Create a server.
     */
    public SDIServer(String defaultConfig, String device, int bufferSize, String notifyWAP)
    {
        defaultConfigurationFileName = defaultConfig;
        deviceName = device;
        inputBufferSize = bufferSize;
        
        if (notifyWAP != null) {
        	notify = notifyWAP;
        }

        /**
         * This thread reads the data and gives the buffers to the
         * FrameSynchronizer object (if it exists), which is the head of the
         * pipeline.
         */
        runner = new Thread(this);
        //runner.setDaemon(true);
        runner.setPriority(Thread.MAX_PRIORITY);
        runner.start();
        
        // assuming for now this is the best place to throw the ingestor restart event
        // we may find out otherwise later...
		try {
    		// create a current date/time string
    		Date rightNow = new Date();
    		String dateTimeString = sdf.format(rightNow); 
    		String [] cmdData = new String[3];
    		cmdData[0] = notify;
    		cmdData[1] = "ingerestart";
    		cmdData[2] = dateTimeString + ":ingestor restarted";
 			Runtime.getRuntime().exec(cmdData);
			// also log the event
			System.out.println(dateTimeString + ":ingestor restarted");
			writeToConsole(dateTimeString + ":ingestor restarted");
		} catch (IOException ioe) {
			System.err.println("ingerestart notification failed, notify binary missing?");
		}        
    }

    /**
     * Load a configuration from a file that is local to the server.
     * @throws IOException 
     * @throws StpsException 
     */
    public void loadFile(String filename) throws StpsException, IOException
    {
        super.loadFile(filename);
        defaultConfigurationFileName = filename;
    }

    /**
     * This thread method gets data from the card and sends it down the
     * pipeline if it exists. It discards the data if the pipeline is null.
     */
    public void run()
    {
        byte[] buffer = new byte[inputBufferSize];
        // we read directly from a FileInputStream to avoid all the method 
        // call overhead associated with BufferedInputStreams.  We can do 
        // this because we are reading fixed-sized chunks in an infinite loop.
        FileInputStream fis = null;

        try {
        	fis = new FileInputStream(deviceName);
        }
        catch (java.io.IOException e) {
        	log("Error reading ingestor card. " + e.getMessage());
        	closeLog();
        	System.exit(-2);
        }
        
        while (running) {
            
            if (fs == null) {
            	
                log("Auto-loading configuration.");
                try
                {
                    if (defaultConfigurationFileName != null)
                    {
                        loadFile(defaultConfigurationFileName);
                    }
                    setEnabled(true);
                }
                catch (StpsException se)
                {
                    closeLog();
                    System.exit(-4);
                }
                catch (IOException ioe)
                {
                    closeLog();
                    System.exit(-4);
                }
            }

            int bytesRead = -1;
            int consecutiveErrors = 0;

            do {	
            	try {
                    bytesRead = fis.read(buffer);
                    // System.err.println("Bytes read from card: " + bytesRead);
                    if (bytesRead < 0) {
                    	consecutiveErrors++;
                    	// System.err.println("Sleeping 2/10 of one second to let data catch up. Consecutive sleep #" + consecutiveErrors);
                    	try {
                    		Thread.sleep(200);
                    	} catch (InterruptedException ie) {
                    		// do nothing
                    	}
                    } else {
                        consecutiveErrors = 0;
                    }
                    if ((bytesRead > 0) && (fs != null))
                    {
                        fs.putBuffer(buffer, bytesRead);
                    }
                    // restart if we had a predefined number of errors in a row
                    if (consecutiveErrors < CONSECUTIVE_READ_ERROR_LIMIT) {
                    	bytesRead = 0;
                    } else {
                    	System.err.println("Exiting read loop, consecutive failed read #" + consecutiveErrors);
                    }
                }
                catch (java.io.IOException ior)
                {
                    log("STPS error reading data socket. I will try again. " + ior.getMessage());
                    ++consecutiveErrors;
                    bytesRead = (consecutiveErrors >= CONSECUTIVE_READ_ERROR_LIMIT)? -1 : 0;
                }
                catch (nasa.stps.core.StpsException se)
                {
                    log("Data read error. " + se.getMessage());
                    se.printStackTrace();
                    bytesRead = -1;
                }
            }
            while (running && (bytesRead >= 0));

            // XXX should this be deleted, or moved outside while loop below?
            //try { 
            //	input.close(); 
            //} catch (java.io.IOException ioe) {
            //	ioe.printStackTrace();
            //}
            try { 
            	if (fs != null) fs.flush(); 
            } catch (nasa.stps.core.StpsException se) {
            	se.printStackTrace();
            }
            state.unload();
            fs = null;
            log("Shutting down the pipeline.");
            
        	// we're going to go back, try and rebuild the pipeline, and start over.
            // pause for a little bit, because if the card is dead this will be a loop...
            running = true;
        	try {
        		Thread.sleep(4000);
        	} catch (InterruptedException ie) {
        		// do nothing
        	}
        }

        log("Server stopping.");
        closeLog();
    }
        
    /**
     * write a message to the system console
     */
    private void writeToConsole(String message) {
        try {
           PrintWriter pw = new PrintWriter(new FileOutputStream("/dev/console"), true);
           pw.println(message);
           pw.close();
        }
        catch (FileNotFoundException fnfe) {
           System.out.println("Cannot write to system console");
        }
    }
    
}
