package nasa.stps.server;

import java.io.*;
import java.net.*;

import nasa.stps.core.StpsException;

/**
 * This is an STPS server that receives telemetry through a TCP/IP socket.
 * <p>
 * You may pass one argument when invoking, the server name. The full server
 * name then becomes "StpsServices."+name. The default full server name is
 * "StpsServices.A".
 * <p>
 * The server understands the following system properties. You set them when
 * you run the server using the -D attribute as in "-Dport=5000".
 * <br>
 * -Dconfig=xmlConfigFileName
 * <br>A configuration file to be used until overridden by a loaded one. The
 * default name is "default.xml".
 * <br>
 * -Dport=4395
 * <br>The port number that the server reads for telemetry data. The
 * default port is 4395.
 * <br>
 * -DbufferSizeKb=8
 * The amount of data I accumulate before processing. The default is 8 kb.
 * <br>
 * -Dsetup=configurationDirectory
 * <br>The directory where local configuration files are found. If provided,
 * all files must be within the directory tree.
 * @author Stan Hilinski, April, 2001
 */
public class TcpServer extends AbstractServer
{
    private int inputBufferSize = 4 * 1024;
    private int inputPort = 4935;

    /**
     * If a data source connects to this server and there is no loaded
     * configuration, the server will automatically load one of the two
     * following configurations. One will be null. When a user loads a
     * new configuration, it also becomes the new automatic default.
     */
    private String defaultConfigurationFileName = null;

    /**
     * This is the main entry point for the STPS server.
     */
    public static void main(String args[])
    {
        String defaultConfig = System.getProperty("config","default.xml");
        int port = 0;
        int bufferSize = 0;
        String propertyName = null;

        try
        {
            //Get the port number system property.
            propertyName = "port";
            String x = System.getProperty(propertyName,"4935");
            port = Integer.valueOf(x).intValue();

            //Get the input buffer size.
            propertyName = "bufferSizeKb";
            x = System.getProperty(propertyName,"8");
            bufferSize = 1024 * Integer.valueOf(x).intValue();
        }
        catch (NumberFormatException nfe)
        {
            System.err.println(propertyName + " property is not an integer.");
            System.exit(-3);
        }

        try
        {
            TcpServer server = new TcpServer(defaultConfig,port,bufferSize);
            server.log("Ready to serve.");
        }
        catch (Exception ex)
        {
            System.err.println("SERVER ERROR. " + ex);
            ex.printStackTrace();
            System.exit(-2);
        }
    }

    /**
     * Create a server.
     */
    public TcpServer(String defaultConfig, int port, int bufferSize)
    {
        defaultConfigurationFileName = defaultConfig;
        inputPort = port;
        inputBufferSize = bufferSize;

        /**
         * This thread reads the sockets and gives the buffers to the
         * FrameSynchronizer object (if it exists), which is the head of the
         * pipeline.
         */
        runner = new Thread(this);
        runner.setDaemon(true);
        runner.setPriority(Thread.MAX_PRIORITY);
        runner.start();
    }

    /**
     * Load a configuration from a file that is local to the server.
     * @throws IOException 
     * @throws StpsException 
     */
    public void loadFile(String filename) throws StpsException, IOException
    {
        super.loadFile(filename);
        defaultConfigurationFileName = filename;
    }

    /**
     * This thread method gets data from the socket and sends it down the
     * pipeline if it exists. It discards the data if the pipeline is null.
     */
    public void run()
    {
        byte[] buffer = new byte[inputBufferSize];
        ServerSocket serverSocket = null;

        try
        {
            //The server socket listens for a data transmission client.
            serverSocket = new ServerSocket(inputPort,0);
        }
        catch (java.io.IOException se)
        {
            log("STPS fatal error creating server socket. "+se.getMessage());
            closeLog();
            System.exit(-1);
        }

        while (running)
        {
            Socket dataSocket = null;
            InputStream input = null;

            try

            {
                dataSocket = serverSocket.accept();
                input = dataSocket.getInputStream();
                input = new BufferedInputStream(input,inputBufferSize);
            }
            catch (java.io.IOException e)
            {
                log("Error reading server socket. "+e.getMessage());
                closeLog();
                System.exit(-2);
            }

            if (fs == null)
            {
                log("Auto-loading default configuration.");
                try
                {
                    if (defaultConfigurationFileName != null)
                    {
                        loadFile(defaultConfigurationFileName);
                    }
                    setEnabled(true);
                }
                catch (Exception e)
                {
                    closeLog();
                    System.exit(-4);
                }
            }

            int bytesRead = -1;
            int consecutiveErrors = 0;

            do
            {
                try
                {
                    bytesRead = input.read(buffer);
                    consecutiveErrors = 0;
                    if ((bytesRead > 0) && (fs != null))
                    {
                        fs.putBuffer(buffer, bytesRead);
                    }
                }
                catch (java.io.IOException ior)
                {
                    log("STPS error reading data socket. I will try again. "+ior.getMessage());
                    ++consecutiveErrors;
                    bytesRead = (consecutiveErrors >= 4)? -1 : 0;
                }
                catch (nasa.stps.core.StpsException se)
                {
                    log("Data read error. "+se.getMessage());
                    bytesRead = -1;
                }
            }
            while (running && (bytesRead >= 0));

            try { input.close(); }
            catch (java.io.IOException e1) {}
            try { dataSocket.close(); }
            catch (java.io.IOException e2) {}
            try { if (fs != null) fs.flush(); }
            catch (nasa.stps.core.StpsException e3) {}
            state.unload();
            fs = null;
            log("I shut down the pipeline.");
        }

        try { serverSocket.close(); }
        catch (java.io.IOException ssio) { }
        log("I stop now.");
        closeLog();
    }
}
